# ===============================================================================
# Copyright 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============================================================================

import numpy as np
from warnings import warn

from mpi4py import MPI
from dpctl import SyclQueue
import dpctl.tensor as dpt
from sklearnex.spmd.linear_model import LinearRegression


def generate_X_y(ns, data_seed):
    nf, nr = 129, 131

    crng = np.random.default_rng(777)
    coef = crng.uniform(-4, 1, size=(nr, nf)).T
    intp = crng.uniform(-1, 9, size=(nr, ))

    drng = np.random.default_rng(data_seed)
    data = drng.uniform(-7, 7, size=(ns, nf))
    resp = data @ coef + intp[np.newaxis, :]

    return data, resp


def get_train_data(rank):
    return generate_X_y(101, rank)


def get_test_data(rank):
    return generate_X_y(1024, rank)


comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()

if size < 2:
    warn("This example was intentionally "
         "designed to run in distributed mode only", RuntimeWarning)

X, y = get_train_data(rank)

queue = SyclQueue("gpu")

dpt_X = dpt.asarray(X, usm_type="device", sycl_queue=queue)
dpt_y = dpt.asarray(y, usm_type="device", sycl_queue=queue)

model = LinearRegression().fit(dpt_X, dpt_y)

print(f"Coefficients on rank {rank}:\n", model.coef_)
print(f"Intercept on rank {rank}:\n", model.intercept_)

X_test, _ = get_test_data(rank)
dpt_X_test = dpt.asarray(X_test, usm_type="device", sycl_queue=queue)

result = model.predict(dpt_X_test)

print(f"Result on rank {rank}:\n", dpt.to_numpy(result))



#===============================================================================
# Copyright 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

import numpy as np
from sklearn.metrics import mean_squared_error
from warnings import warn
from mpi4py import MPI
import dpctl
import dpctl.tensor as dpt
from numpy.testing import assert_allclose
from sklearnex.spmd.neighbors import KNeighborsRegressor


def generate_X_y(par, coef_seed, data_seed):
    ns, nf = par['ns'], par['nf']

    crng = np.random.default_rng(coef_seed)
    coef = crng.uniform(-10, 10, size=(nf,))

    drng = np.random.default_rng(data_seed)
    data = drng.uniform(-100, 100, size=(ns, nf))
    resp = data @ coef

    return data, resp, coef


comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()

if dpctl.has_gpu_devices:
    q = dpctl.SyclQueue("gpu")
else:
    raise RuntimeError('GPU devices unavailable. Currently, '
          'SPMD execution mode is implemented only for this device type.')

params_train = {'ns': 1000000, 'nf': 3}
params_test = {'ns': 100, 'nf': 3}

X_train, y_train, coef_train = generate_X_y(params_train, 10, rank)
X_test, y_test, coef_test = generate_X_y(params_test, 10, rank + 99)

dpt_X_train = dpt.asarray(X_train, usm_type="device", sycl_queue=q)
dpt_y_train = dpt.asarray(y_train, usm_type="device", sycl_queue=q)
dpt_X_test = dpt.asarray(X_test, usm_type="device", sycl_queue=q)
# dpt_y_test = dpt.asarray(y_test, usm_type="device", sycl_queue=q)

assert_allclose(coef_train, coef_test)

model_spmd = KNeighborsRegressor(algorithm='brute',
                                 n_neighbors=5,
                                 weights='uniform',
                                 p=2,
                                 metric='minkowski')
model_spmd.fit(dpt_X_train, dpt_y_train)

y_predict = model_spmd.predict(dpt_X_test)

print("Brute Force Distributed kNN regression results:")
print("Ground truth (first 5 observations on rank {}):\n{}".format(rank, y_test[:5]))
print("Regression results (first 5 observations on rank {}):\n{}"
      .format(rank, dpt.to_numpy(y_predict)[:5]))
print("RMSE for entire rank {}: {}\n"
      .format(rank, mean_squared_error(y_test, dpt.to_numpy(y_predict), squared=False)))



#===============================================================================
# Copyright 2014 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

import numpy as np
import numbers
from math import sqrt
from scipy.sparse import issparse

from sklearn.utils import check_array
from sklearn.utils.validation import check_is_fitted
from sklearn.utils.extmath import stable_cumsum

import daal4py
from .._utils import (
    getFPType, sklearn_check_version, PatchingConditionsChain)
from .._device_offload import support_usm_ndarray

if sklearn_check_version('0.22'):
    from sklearn.decomposition._pca import PCA as PCA_original
else:
    from sklearn.decomposition.pca import PCA as PCA_original

if sklearn_check_version('0.23'):
    from sklearn.decomposition._pca import _infer_dimension
elif sklearn_check_version('0.22'):
    from sklearn.decomposition._pca import _infer_dimension_
else:
    from sklearn.decomposition.pca import _infer_dimension_


class PCA(PCA_original):
    __doc__ = PCA_original.__doc__

    def __init__(
        self,
        n_components=None,
        copy=True,
        whiten=False,
        svd_solver='auto',
        tol=0.0,
        iterated_power='auto',
        n_oversamples=10,
        power_iteration_normalizer="auto",
        random_state=None
    ):
        self.n_components = n_components
        self.copy = copy
        self.whiten = whiten
        self.svd_solver = svd_solver
        self.tol = tol
        self.iterated_power = iterated_power
        self.n_oversamples = n_oversamples
        self.power_iteration_normalizer = power_iteration_normalizer
        self.random_state = random_state

    def _validate_n_components(self, n_components, n_samples, n_features):
        if n_components == 'mle':
            if n_samples < n_features:
                raise ValueError("n_components='mle' is only supported "
                                 "if n_samples >= n_features")
        elif not 0 <= n_components <= min(n_samples, n_features):
            raise ValueError("n_components=%r must be between 0 and "
                             "min(n_samples, n_features)=%r with "
                             "svd_solver='full'"
                             % (n_components, min(n_samples, n_features)))
        elif n_components >= 1:
            if not isinstance(n_components, numbers.Integral):
                raise ValueError("n_components=%r must be of type int "
                                 "when greater than or equal to 1, "
                                 "was of type=%r"
                                 % (n_components, type(n_components)))

    def _fit_full_daal4py(self, X, n_components):
        n_samples, n_features = X.shape
        n_sf_min = min(n_samples, n_features)

        if n_components == 'mle':
            daal_n_components = n_features
        elif n_components < 1:
            daal_n_components = n_sf_min
        else:
            daal_n_components = n_components

        fpType = getFPType(X)

        covariance_algo = daal4py.covariance(
            fptype=fpType, outputMatrixType='covarianceMatrix')
        covariance_res = covariance_algo.compute(X)

        self.mean_ = covariance_res.mean.ravel()
        covariance = covariance_res.covariance
        variances_ = np.array([covariance[i, i] for i in range(n_features)])

        pca_alg = daal4py.pca(
            fptype=fpType,
            method='correlationDense',
            resultsToCompute='eigenvalue',
            isDeterministic=True,
            nComponents=daal_n_components
        )
        pca_res = pca_alg.compute(X, covariance)

        components_ = pca_res.eigenvectors
        explained_variance_ = np.maximum(pca_res.eigenvalues.ravel(), 0)
        tot_var = explained_variance_.sum()
        explained_variance_ratio_ = explained_variance_ / tot_var

        if n_components == 'mle':
            if sklearn_check_version('0.23'):
                n_components = _infer_dimension(explained_variance_, n_samples)
            else:
                n_components = \
                    _infer_dimension_(explained_variance_, n_samples, n_features)
        elif 0 < n_components < 1.0:
            ratio_cumsum = stable_cumsum(explained_variance_ratio_)
            n_components = np.searchsorted(ratio_cumsum, n_components,
                                           side='right') + 1

        if n_components < n_sf_min:
            if explained_variance_.shape[0] == n_sf_min:
                self.noise_variance_ = explained_variance_[n_components:].mean()
            else:
                resid_var_ = variances_.sum()
                resid_var_ -= explained_variance_[:n_components].sum()
                self.noise_variance_ = resid_var_ / (n_sf_min - n_components)
        else:
            self.noise_variance_ = 0.

        if sklearn_check_version('1.2'):
            self.n_samples_, self.n_features_in_ = n_samples, n_features
        else:
            self.n_samples_, self.n_features_ = n_samples, n_features
        self.components_ = components_[:n_components]
        self.n_components_ = n_components
        self.explained_variance_ = explained_variance_[:n_components]
        self.explained_variance_ratio_ = explained_variance_ratio_[:n_components]
        self.singular_values_ = np.sqrt((n_samples - 1) * self.explained_variance_)

    def _fit_full(self, X, n_components):
        n_samples, n_features = X.shape
        self._validate_n_components(n_components, n_samples, n_features)

        self._fit_full_daal4py(X, min(X.shape))

        U = None
        V = self.components_
        S = self.singular_values_

        if n_components == 'mle':
            if sklearn_check_version('0.23'):
                n_components = _infer_dimension(self.explained_variance_, n_samples)
            else:
                n_components = \
                    _infer_dimension_(self.explained_variance_, n_samples, n_features)
        elif 0 < n_components < 1.0:
            ratio_cumsum = stable_cumsum(self.explained_variance_ratio_)
            n_components = np.searchsorted(ratio_cumsum, n_components,
                                           side='right') + 1

        if n_components < min(n_features, n_samples):
            self.noise_variance_ = self.explained_variance_[n_components:].mean()
        else:
            self.noise_variance_ = 0.

        if sklearn_check_version('1.2'):
            self.n_samples_, self.n_features_in_ = n_samples, n_features
        else:
            self.n_samples_, self.n_features_ = n_samples, n_features
        self.components_ = self.components_[:n_components]
        self.n_components_ = n_components
        self.explained_variance_ = self.explained_variance_[:n_components]
        self.explained_variance_ratio_ = self.explained_variance_ratio_[:n_components]
        self.singular_values_ = self.singular_values_[:n_components]

        return U, S, V

    def _fit(self, X):
        if issparse(X):
            raise TypeError('PCA does not support sparse input. See '
                            'TruncatedSVD for a possible alternative.')

        if sklearn_check_version('0.23'):
            X = self._validate_data(X, dtype=[np.float64, np.float32],
                                    ensure_2d=True, copy=False)
        else:
            X = check_array(X, dtype=[np.float64, np.float32], ensure_2d=True, copy=False)

        if self.n_components is None:
            if self.svd_solver != 'arpack':
                n_components = min(X.shape)
            else:
                n_components = min(X.shape) - 1
        else:
            n_components = self.n_components

        self._fit_svd_solver = self.svd_solver
        shape_good_for_daal = X.shape[1] / X.shape[0] < 2

        if self._fit_svd_solver == 'auto':
            if sklearn_check_version('1.1'):
                # Small problem or n_components == 'mle', just call full PCA
                if max(X.shape) <= 500 or n_components == "mle":
                    self._fit_svd_solver = "full"
                elif 1 <= n_components < 0.8 * min(X.shape):
                    self._fit_svd_solver = "randomized"
                # This is also the case of n_components in (0,1)
                else:
                    self._fit_svd_solver = "full"
            else:
                if n_components == 'mle':
                    self._fit_svd_solver = 'full'
                else:
                    n, p, k = X.shape[0], X.shape[1], n_components
                    # These coefficients are result of training of Logistic Regression
                    # (max_iter=10000, solver="liblinear", fit_intercept=False)
                    # on different datasets and number of components.
                    # X is a dataset with npk, np^2, and n^2 columns.
                    # And y is speedup of patched scikit-learn's
                    # full PCA against stock scikit-learn's randomized PCA.
                    regression_coefs = np.array([
                        [9.779873e-11, n * p * k],
                        [-1.122062e-11, n * p * p],
                        [1.127905e-09, n ** 2],
                    ])

                    if n_components >= 1 and np.dot(
                            regression_coefs[:, 0], regression_coefs[:, 1]) <= 0:
                        self._fit_svd_solver = 'randomized'
                    else:
                        self._fit_svd_solver = 'full'

        if not shape_good_for_daal or self._fit_svd_solver != 'full':
            if sklearn_check_version('0.23'):
                X = self._validate_data(X, copy=self.copy)
            else:
                X = check_array(X, copy=self.copy)

        _patching_status = PatchingConditionsChain(
            "sklearn.decomposition.PCA.fit")
        _dal_ready = _patching_status.and_conditions([
            (self._fit_svd_solver == 'full',
                f"'{self._fit_svd_solver}' SVD solver is not supported. "
                "Only 'full' solver is supported.")
        ])

        if _dal_ready:
            _dal_ready = _patching_status.and_conditions([
                (shape_good_for_daal,
                    "The shape of X does not satisfy oneDAL requirements: "
                    "number of features / number of samples >= 2")
            ])
            if _dal_ready:
                result = self._fit_full(X, n_components)
            else:
                result = PCA_original._fit_full(self, X, n_components)
        elif self._fit_svd_solver in ['arpack', 'randomized']:
            result = self._fit_truncated(X, n_components, self._fit_svd_solver)
        else:
            raise ValueError("Unrecognized svd_solver='{0}'"
                             "".format(self._fit_svd_solver))

        _patching_status.write_log()
        return result

    def _transform_daal4py(self, X, whiten=False, scale_eigenvalues=True, check_X=True):
        if sklearn_check_version('0.22'):
            check_is_fitted(self)
        else:
            check_is_fitted(self, ['mean_', 'components_'], all_or_any=all)

        if sklearn_check_version("1.0"):
            self._check_feature_names(X, reset=False)
        X = check_array(X, dtype=[np.float64, np.float32], force_all_finite=check_X)
        fpType = getFPType(X)

        tr_data = dict()
        if self.mean_ is not None:
            tr_data['mean'] = self.mean_.reshape((1, -1))
        if whiten:
            if scale_eigenvalues:
                tr_data['eigenvalue'] = \
                    (self.n_samples_ - 1) * self.explained_variance_.reshape((1, -1))
            else:
                tr_data['eigenvalue'] = self.explained_variance_.reshape((1, -1))
        elif scale_eigenvalues:
            tr_data['eigenvalue'] = np.full(
                (1, self.explained_variance_.shape[0]),
                self.n_samples_ - 1.0, dtype=X.dtype)

        if sklearn_check_version('1.2'):
            expected_n_features = self.n_features_in_
        else:
            expected_n_features = self.n_features_
        if X.shape[1] != expected_n_features:
            raise ValueError(
                (f'X has {X.shape[1]} features, '
                 f'but PCA is expecting {expected_n_features} features as input'))

        tr_res = daal4py.pca_transform(
            fptype=fpType
        ).compute(X, self.components_, tr_data)

        return tr_res.transformedData

    @support_usm_ndarray()
    def transform(self, X):
        """
        Apply dimensionality reduction to X.

        X is projected on the first principal components previously extracted
        from a training set.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            New data, where `n_samples` is the number of samples
            and `n_features` is the number of features.

        Returns
        -------
        X_new : array-like of shape (n_samples, n_components)
            Projection of X in the first principal components, where `n_samples`
            is the number of samples and `n_components` is the number of the components.
        """
        _patching_status = PatchingConditionsChain(
            "sklearn.decomposition.PCA.transform")
        _dal_ready = _patching_status.and_conditions([
            (self.n_components_ > 0, "Number of components <= 0.")
        ])

        _patching_status.write_log()
        if _dal_ready:
            return self._transform_daal4py(X, whiten=self.whiten,
                                           check_X=True, scale_eigenvalues=False)
        return PCA_original.transform(self, X)

    @support_usm_ndarray()
    def fit_transform(self, X, y=None):
        """
        Fit the model with X and apply the dimensionality reduction on X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data, where `n_samples` is the number of samples
            and `n_features` is the number of features.

        y : Ignored
            Ignored.

        Returns
        -------
        X_new : ndarray of shape (n_samples, n_components)
            Transformed values.

        Notes
        -----
        This method returns a Fortran-ordered array. To convert it to a
        C-ordered array, use 'np.ascontiguousarray'.
        """

        if sklearn_check_version('1.2'):
            self._validate_params()

        U, S, Vt = self._fit(X)

        _patching_status = PatchingConditionsChain(
            "sklearn.decomposition.PCA.fit_transform")
        _dal_ready = _patching_status.and_conditions([
            (U is None, "Stock fitting was used.")
        ])
        if _dal_ready:
            _dal_ready = _patching_status.and_conditions([
                (self.n_components_ > 0, "Number of components <= 0.")
            ])
            if _dal_ready:
                result = self._transform_daal4py(
                    X, whiten=self.whiten, check_X=False, scale_eigenvalues=False)
            else:
                result = np.empty((self.n_samples_, 0), dtype=X.dtype)
        else:
            U = U[:, :self.n_components_]

            if self.whiten:
                U *= sqrt(X.shape[0] - 1)
            else:
                U *= S[:self.n_components_]

            result = U

        _patching_status.write_log()
        return result



#===============================================================================
# Copyright 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

# Calling scikit-learn patch - this would enable acceleration on all
# enabled algorithms. This is most straight forward way of patching
from sklearnex import patch_sklearn
patch_sklearn()

# Function that can validate current state of patching
from sklearnex import sklearn_is_patched
sklearn_is_patched()

# Calling scikit-learn unpatch - this would revert patching for all algorithms
from sklearnex import unpatch_sklearn
unpatch_sklearn()

# Direct import of functions in way aligned with scikit-learn
from sklearnex.neighbors import NearestNeighbors

#Patching can be enabled for selected algorithms/estimators only
patch_sklearn(["DBSCAN"])

#Remaining non modified scikit-learn codes
from sklearn.datasets import make_blobs
from sklearn.preprocessing import StandardScaler

centers = [[1, 1], [-1, -1], [1, -1]]
X, labels_true = make_blobs(
    n_samples=750, centers=centers, cluster_std=0.4, random_state=0
)

X = StandardScaler().fit_transform(X)

from sklearn.cluster import DBSCAN

db = DBSCAN(eps=0.3, min_samples=10).fit(X)
labels = db.labels_

# Number of clusters in labels, ignoring noise if present.
n_clusters_ = len(set(labels)) - (1 if -1 in labels else 0)
n_noise_ = list(labels).count(-1)

print("Estimated number of clusters: %d" % n_clusters_)
print("Estimated number of noise points: %d" % n_noise_)



# ===============================================================================
# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============================================================================

from .kernel_functions import linear_kernel, rbf_kernel, poly_kernel, sigmoid_kernel
from .get_tree import get_tree_state_cls, get_tree_state_reg

__all__ = [
    'get_tree_state_cls',
    'get_tree_state_reg',
    'linear_kernel',
    'rbf_kernel',
    'poly_kernel',
    'sigmoid_kernel']



#===============================================================================
# Copyright 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

import numpy as np

from onedal import _backend
from ..common._policy import _get_policy
from ..datatypes._data_conversion import from_table, to_table
from ..datatypes import _convert_to_supported
from daal4py.sklearn._utils import sklearn_check_version


class PCA():
    def __init__(
        self,
        n_components=None,
        is_deterministic=True,
        method='precomputed',
        copy=True
    ):
        self.n_components = n_components
        self.method = method
        self.is_deterministic = is_deterministic

    def get_onedal_params(self, data):
        return {
            'fptype':
                'float' if data.dtype == np.float32 else 'double',
            'method': self.method,
            'n_components': self.n_components,
            'is_deterministic': self.is_deterministic
        }

    def _get_policy(self, queue, *data):
        return _get_policy(queue, *data)

    def fit(self, X, queue):
        n_samples, n_features = X.shape
        n_sf_min = min(n_samples, n_features)

        policy = self._get_policy(queue, X)
        # TODO: investigate why np.ndarray with OWNDATA=FALSE flag
        # fails to be converted to oneDAL table
        if isinstance(X, np.ndarray) and not X.flags['OWNDATA']:
            X = X.copy()
        X = _convert_to_supported(policy, X)

        params = self.get_onedal_params(X)
        cov_result = _backend.covariance.compute(
            policy,
            {'fptype': params['fptype'], 'method': 'dense'},
            to_table(X)
        )
        covariance_matrix = from_table(cov_result.cov_matrix)
        self.mean_ = from_table(cov_result.means)
        result = _backend.decomposition.dim_reduction.train(
            policy,
            params,
            to_table(covariance_matrix)
        )

        self.n_components_ = self.n_components
        self.variances_ = from_table(result.variances)
        self.components_ = from_table(result.eigenvectors)
        self.explained_variance_ = \
            np.maximum(from_table(result.eigenvalues).ravel(), 0)
        tot_var = covariance_matrix.trace()
        self.explained_variance_ratio_ = self.explained_variance_ / tot_var
        self.singular_values_ = np.sqrt(
            (n_samples - 1) * self.explained_variance_
        )

        if sklearn_check_version("1.2"):
            self.n_features_in_ = n_features
        elif sklearn_check_version("0.24"):
            self.n_features_in_ = n_features
            self.n_features_ = n_features
        else:
            self.n_features_ = n_features

        self.n_samples_ = n_samples
        if self.n_components < n_sf_min:
            if self.explained_variance_.shape[0] < n_sf_min:
                resid_var_ = tot_var - \
                    self.explained_variance_[:self.n_components].sum()
                self.noise_variance_ = \
                    resid_var_ / (n_sf_min - self.n_components)
        return self

    def _create_model(self):
        m = _backend.decomposition.dim_reduction.model()
        m.eigenvectors = to_table(self.components_)
        self._onedal_model = m
        return m

    def predict(self, X, queue):
        policy = self._get_policy(queue, X)
        model = self._create_model()

        X = _convert_to_supported(policy, X)
        params = self.get_onedal_params(X)
        result = _backend.decomposition.dim_reduction.infer(policy,
                                                            params,
                                                            model,
                                                            to_table(X))
        return from_table(result.transformed_data)



# ===============================================================================
# Copyright 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============================================================================

from daal4py.sklearn._utils import daal_check_version, sklearn_check_version

if daal_check_version((2023, 'P', 100)):
    import pytest
    import numpy as np
    from numpy.testing import assert_allclose, assert_array_equal

    from onedal.linear_model import LinearRegression
    from onedal.tests.utils._device_selection import get_queues

    from sklearn.datasets import load_diabetes
    from sklearn.metrics import mean_squared_error
    from sklearn.model_selection import train_test_split

    @pytest.mark.parametrize('queue', get_queues())
    @pytest.mark.parametrize('dtype', [np.float32, np.float64])
    def test_diabetes(queue, dtype):
        X, y = load_diabetes(return_X_y=True)
        X, y = X.astype(dtype), y.astype(dtype)
        X_train, X_test, y_train, y_test = \
            train_test_split(X, y,
                             train_size=0.8, random_state=777)
        model = LinearRegression(fit_intercept=True)
        model.fit(X_train, y_train, queue=queue)
        y_pred = model.predict(X_test, queue=queue)
        assert mean_squared_error(y_test, y_pred) < 2396

    @pytest.mark.parametrize('queue', get_queues())
    @pytest.mark.parametrize('dtype', [np.float32, np.float64])
    def test_pickle(queue, dtype):
        X, y = load_diabetes(return_X_y=True)
        X, y = X.astype(dtype), y.astype(dtype)
        model = LinearRegression(fit_intercept=True)
        model.fit(X, y, queue=queue)
        expected = model.predict(X, queue=queue)

        import pickle
        dump = pickle.dumps(model)
        model2 = pickle.loads(dump)

        assert isinstance(model2, model.__class__)
        result = model2.predict(X, queue=queue)

        assert_array_equal(expected, result)

    @pytest.mark.parametrize('queue', get_queues())
    @pytest.mark.parametrize('dtype', [np.float32, np.float64])
    def test_full_results(queue, dtype):
        seed = 42
        f_count, r_count = 19, 7
        s_count, t_count = 3500, 1999

        gen = np.random.default_rng(seed)
        intp = gen.random(size=r_count, dtype=dtype)
        coef = gen.random(size=(r_count, f_count), dtype=dtype).T

        X = gen.random(size=(s_count, f_count), dtype=dtype)
        y = X @ coef + intp[np.newaxis, :]

        model = LinearRegression(fit_intercept=True)
        model.fit(X, y, queue=queue)

        tol = 2e-3 if model.coef_.dtype == np.float32 else 1e-5
        assert_allclose(coef, model.coef_.T, rtol=tol)

        tol = 2e-3 if model.intercept_.dtype == np.float32 else 1e-5
        assert_allclose(intp, model.intercept_, rtol=tol)

        Xt = gen.random(size=(t_count, f_count), dtype=dtype)
        gtr = Xt @ coef + intp[np.newaxis, :]

        res = model.predict(Xt, queue=queue)

        tol = 2e-5 if res.dtype == np.float32 else 1e-7
        assert_allclose(gtr, res, rtol=tol)

    @pytest.mark.parametrize('queue', get_queues())
    @pytest.mark.parametrize('dtype', [np.float32, np.float64])
    def test_no_intercept_results(queue, dtype):
        seed = 42
        f_count, r_count = 19, 7
        s_count, t_count = 3500, 1999

        gen = np.random.default_rng(seed)
        coef = gen.random(size=(r_count, f_count), dtype=dtype).T

        X = gen.random(size=(s_count, f_count), dtype=dtype)
        y = X @ coef

        model = LinearRegression(fit_intercept=False)
        model.fit(X, y, queue=queue)

        tol = 2e-3 if model.coef_.dtype == np.float32 else 1e-7
        assert_allclose(coef, model.coef_.T, rtol=tol)

        Xt = gen.random(size=(t_count, f_count), dtype=dtype)
        gtr = Xt @ coef

        res = model.predict(Xt, queue=queue)

        tol = 5e-5 if res.dtype == np.float32 else 1e-7
        assert_allclose(gtr, res, rtol=tol)

    @pytest.mark.parametrize('queue', get_queues())
    @pytest.mark.parametrize('dtype', [np.float32, np.float64])
    def test_reconstruct_model(queue, dtype):
        seed = 42
        s_count = 3500
        f_count, r_count = 14, 9

        gen = np.random.default_rng(seed)
        intp = gen.random(size=r_count, dtype=dtype)
        coef = gen.random(size=(r_count, f_count), dtype=dtype).T

        X = gen.random(size=(s_count, f_count), dtype=dtype)
        gtr = X @ coef + intp[np.newaxis, :]

        model = LinearRegression(fit_intercept=True)
        model.coef_ = coef.T
        model.intercept_ = intp

        res = model.predict(X, queue=queue)

        tol = 1e-5 if res.dtype == np.float32 else 1e-7
        assert_allclose(gtr, res, rtol=tol)



#===============================================================================
# Copyright 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

# sklearnex RF example for distributed systems; SPMD mode
# run like this:
#    mpirun -n 4 python ./random_forest_regressor_spmd.py

import numpy as np

import dpctl
import dpctl.tensor as dpt

from mpi4py import MPI
from sklearnex.spmd.ensemble import RandomForestRegressor

from numpy.testing import assert_allclose


def generate_X_y(par, coef_seed, data_seed):
    ns, nf = par['ns'], par['nf']

    crng = np.random.default_rng(coef_seed)
    coef = crng.uniform(-10, 10, size=(nf,))

    drng = np.random.default_rng(data_seed)
    data = drng.uniform(-100, 100, size=(ns, nf))
    resp = data @ coef

    return data, resp, coef


comm = MPI.COMM_WORLD
mpi_size = comm.Get_size()
mpi_rank = comm.Get_rank()

params_train = {'ns': 10000, 'nf': 3}
params_test = {'ns': 100, 'nf': 3}

X_train, y_train, coef_train = generate_X_y(params_train, 10, mpi_rank)
X_test, y_test, coef_test = generate_X_y(params_test, 10, mpi_rank + 99)

assert_allclose(coef_train, coef_test)

q = dpctl.SyclQueue("gpu")  # GPU

dpt_X_train = dpt.asarray(X_train, usm_type="device", sycl_queue=q)
dpt_y_train = dpt.asarray(y_train, usm_type="device", sycl_queue=q)
dpt_X_test = dpt.asarray(X_test, usm_type="device", sycl_queue=q)

rf = RandomForestRegressor(max_depth=2, random_state=0).fit(dpt_X_train, dpt_y_train)

y_predict = rf.predict(dpt_X_test)

print("Ground truth (first 5 observations on rank {}):\n{}".format(mpi_rank, y_test[:5]))
print("Regression results (first 5 observations on rank {}):\n{}"
      .format(mpi_rank, dpt.to_numpy(y_predict)[:5]))



#===============================================================================
# Copyright 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================


from ...common._spmd_policy import _get_spmd_policy
from ..._device_offload import support_usm_ndarray
from onedal.decomposition.pca import PCA as PCABatch


class BasePCASPMD:
    def _get_policy(self, queue, *data):
        return _get_spmd_policy(queue)


class PCA(BasePCASPMD, PCABatch):

    @support_usm_ndarray()
    def fit(self, X, queue):
        return super().fit(X, queue)

    @support_usm_ndarray()
    def predict(self, X, queue):
        return super().predict(X, queue)



# ===============================================================================
# Copyright 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============================================================================

import numpy as np
from mpi4py import MPI

from dpctl import SyclQueue
import dpctl.tensor as dpt
from sklearnex.spmd.basic_statistics import BasicStatistics as BasicStatisticsSpmd


def generate_data(par, size, seed=777):
    ns, nf = par['ns'], par['nf']

    data_blocks, weight_blocks = [], []
    rng = np.random.default_rng(seed)

    for b in range(size):
        data = rng.uniform(b, (b + 1) * (b + 1),
                           size=(ns, nf))
        weights = rng.uniform(1, (b + 1), size=ns)
        weight_blocks.append(weights)
        data_blocks.append(data)

    data = np.concatenate(data_blocks, axis=0)
    weights = np.concatenate(weight_blocks)

    return (data, weights)


q = SyclQueue("gpu")

comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()

params_spmd = {'ns': 19, 'nf': 31}

data, weights = generate_data(params_spmd, size)
weighted_data = np.diag(weights) @ data

dpt_data = dpt.asarray(data, usm_type="device", sycl_queue=q)
dpt_weights = dpt.asarray(weights, usm_type="device", sycl_queue=q)

gtr_mean = np.mean(weighted_data, axis=0)
gtr_std = np.std(weighted_data, axis=0)

bss = BasicStatisticsSpmd(["mean", "standard_deviation"])
res = bss.compute(dpt_data, dpt_weights)

print(f"Computed mean on rank {rank}:\n", res["mean"])
print(f"Computed std on rank {rank}:\n", res["standard_deviation"])



# ===============================================================================
# Copyright 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============================================================================

import pytest
import numpy as np
from numpy.testing import assert_allclose

from daal4py.sklearn._utils import daal_check_version
from onedal.ensemble import RandomForestClassifier, RandomForestRegressor
from onedal.tests.utils._device_selection import get_queues

from sklearn.datasets import make_classification, make_regression


@pytest.mark.parametrize('queue', get_queues())
def test_rf_classifier(queue):
    X, y = make_classification(n_samples=100, n_features=4,
                               n_informative=2, n_redundant=0,
                               random_state=0, shuffle=False)
    rf = RandomForestClassifier(
        max_depth=2, random_state=0).fit(X, y, queue=queue)
    assert_allclose([1], rf.predict([[0, 0, 0, 0]], queue=queue))


@pytest.mark.parametrize('queue', get_queues())
def test_rf_regression(queue):
    X, y = make_regression(n_samples=100, n_features=4, n_informative=2,
                           random_state=0, shuffle=False)
    rf = RandomForestRegressor(
        max_depth=2, random_state=0).fit(X, y, queue=queue)
    assert_allclose(
        [-6.83], rf.predict([[0, 0, 0, 0]], queue=queue), atol=1e-2)


@pytest.mark.skipif(not daal_check_version((2023, 'P', 101)),
                    reason='requires OneDAL 2023.1.1')
@pytest.mark.parametrize('queue', get_queues('gpu'))
def test_rf_classifier_random_splitter(queue):
    X, y = make_classification(n_samples=100, n_features=4,
                               n_informative=2, n_redundant=0,
                               random_state=0, shuffle=False)
    rf = RandomForestClassifier(
        max_depth=2, random_state=0,
        splitter_mode='random').fit(X, y, queue=queue)
    assert_allclose([1], rf.predict([[0, 0, 0, 0]], queue=queue))


@pytest.mark.parametrize('queue', get_queues('gpu'))
def test_rf_regression_random_splitter(queue):
    X, y = make_regression(n_samples=100, n_features=4, n_informative=2,
                           random_state=0, shuffle=False)
    rf = RandomForestRegressor(
        max_depth=2, random_state=0,
        splitter_mode='random').fit(X, y, queue=queue)
    assert_allclose(
        [-6.83], rf.predict([[0, 0, 0, 0]], queue=queue), atol=1e-2)



#===============================================================================
# Copyright 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

from abc import ABC
from ...common._spmd_policy import _get_spmd_policy
from ..._device_offload import support_usm_ndarray
from onedal.neighbors import KNeighborsClassifier as KNeighborsClassifier_Batch
from onedal.neighbors import KNeighborsRegressor as KNeighborsRegressor_Batch


class NeighborsCommonBaseSPMD(ABC):
    def _get_policy(self, queue, *data):
        return _get_spmd_policy(queue)


class KNeighborsClassifier(NeighborsCommonBaseSPMD, KNeighborsClassifier_Batch):

    @support_usm_ndarray()
    def fit(self, X, y, queue=None):
        return super().fit(X, y, queue)

    @support_usm_ndarray()
    def predict(self, X, queue=None):
        return super().predict(X, queue)

    @support_usm_ndarray()
    def predict_proba(self, X, queue=None):
        raise NotImplementedError("predict_proba not supported in distributed mode.")

    @support_usm_ndarray()
    def kneighbors(self, X=None, n_neighbors=None,
                   return_distance=True, queue=None):
        return super().kneighbors(X, n_neighbors, return_distance, queue)


class KNeighborsRegressor(NeighborsCommonBaseSPMD, KNeighborsRegressor_Batch):
    @support_usm_ndarray()
    def fit(self, X, y, queue=None):
        if queue is not None and queue.sycl_device.is_gpu:
            return super()._fit(X, y, queue=queue)
        else:
            raise ValueError('SPMD version of kNN is not implemented for '
                  'CPU. Consider running on it on GPU.')

    @support_usm_ndarray()
    def kneighbors(self, X=None, n_neighbors=None,
                   return_distance=True, queue=None):
        return super().kneighbors(X, n_neighbors, return_distance, queue)

    @support_usm_ndarray()
    def predict(self, X, queue=None):
        return self._predict_gpu(X, queue=queue)

    def _get_onedal_params(self, X, y=None):
        params = super()._get_onedal_params(X, y)
        if 'responses' not in params['result_option']:
            params['result_option'] += '|responses'
        return params


class NearestNeighbors(NeighborsCommonBaseSPMD):

    @support_usm_ndarray()
    def fit(self, X, y, queue=None):
        return super().fit(X, y, queue)

    @support_usm_ndarray()
    def kneighbors(self, X=None, n_neighbors=None,
                   return_distance=True, queue=None):
        return super().kneighbors(X, n_neighbors, return_distance, queue)



#===============================================================================
# Copyright 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

from onedal import _backend, _is_dpc_backend
import sys

if _is_dpc_backend:
    class _SPMDDataParallelInteropPolicy(_backend.spmd_data_parallel_policy):
        def __init__(self, queue):
            self._queue = queue
            super().__init__(self._queue)

    def _get_spmd_policy(queue):
        # TODO:
        # cases when queue is None
        return _SPMDDataParallelInteropPolicy(queue)



# ===============================================================================
# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============================================================================

import numpy as np
import warnings

from onedal import _is_dpc_backend
from onedal import _backend
from daal4py.sklearn._utils import make2d

try:
    import dpctl
    import dpctl.tensor as dpt
    dpctl_available = dpctl.__version__ >= '0.14'
except ImportError:
    dpctl_available = False


def _apply_and_pass(func, *args):
    if len(args) == 1:
        return func(args[0])
    return tuple(map(func, args))


def from_table(*args):
    return _apply_and_pass(_backend.from_table, *args)


def convert_one_to_table(arg):
    if dpctl_available:
        if isinstance(arg, dpt.usm_ndarray):
            return _backend.dpctl_to_table(arg)
    arg = make2d(arg)
    return _backend.to_table(arg)


def to_table(*args):
    return _apply_and_pass(convert_one_to_table, *args)


if _is_dpc_backend:
    from ..common._policy import _HostInteropPolicy

    def _convert_to_supported(policy, *data):
        def func(x):
            return x

        # CPUs support FP64 by default
        if isinstance(policy, _HostInteropPolicy):
            return _apply_and_pass(func, *data)

        # It can be either SPMD or DPCPP policy
        device = policy._queue.sycl_device

        def convert_or_pass(x):
            if (x is not None) and (x.dtype == np.float64):
                warnings.warn("Data will be converted into float32 from "
                              "float64 because device does not support it",
                              RuntimeWarning, )
                return x.astype(np.float32)
            else:
                return x

        if not device.has_aspect_fp64:
            func = convert_or_pass

        return _apply_and_pass(func, *data)

else:
    def _convert_to_supported(policy, *data):
        def func(x):
            return x

        return _apply_and_pass(func, *data)


#TODO: Update with dpctl
#For now it will fall back to numpy
def _convert_one_to_dataframe(policy, x):
    is_numpy = isinstance(x, np.ndarray)
    if (x is None) or is_numpy:
        return x
    else:
        return np.asarray(x)


def _convert_to_dataframe(policy, *data):
    def _convert_one(x):
        return _convert_one_to_dataframe(policy, x)
    return _apply_and_pass(_convert_one, *data)



#===============================================================================
# Copyright 2014 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

import numpy as np
from scipy import sparse as sp
import numbers

from sklearn.utils import check_array
from sklearn.utils.validation import _check_sample_weight

from sklearn.cluster import DBSCAN as DBSCAN_original

import daal4py
from daal4py.sklearn._utils import (
    make2d, getFPType, PatchingConditionsChain)

from .._device_offload import support_usm_ndarray
from .._utils import sklearn_check_version

if sklearn_check_version('1.1') and not sklearn_check_version('1.2'):
    from sklearn.utils import check_scalar


def _daal_dbscan(X, eps=0.5, min_samples=5, sample_weight=None):
    ww = make2d(sample_weight) if sample_weight is not None else None
    XX = make2d(X)

    fpt = getFPType(XX)
    alg = daal4py.dbscan(
        method='defaultDense',
        fptype=fpt,
        epsilon=float(eps),
        minObservations=int(min_samples),
        memorySavingMode=False,
        resultsToCompute="computeCoreIndices"
    )

    daal_res = alg.compute(XX, ww)
    assignments = daal_res.assignments.ravel()
    if daal_res.coreIndices is not None:
        core_ind = daal_res.coreIndices.ravel()
    else:
        core_ind = np.array([], dtype=np.intc)

    return (core_ind, assignments)


class DBSCAN(DBSCAN_original):
    """Perform DBSCAN clustering from vector array or distance matrix.

    DBSCAN - Density-Based Spatial Clustering of Applications with Noise.
    Finds core samples of high density and expands clusters from them.
    Good for data which contains clusters of similar density.

    Read more in the :ref:`User Guide <dbscan>`.

    Parameters
    ----------
    eps : float, optional
        The maximum distance between two samples for one to be considered
        as in the neighborhood of the other. This is not a maximum bound
        on the distances of points within a cluster. This is the most
        important DBSCAN parameter to choose appropriately for your data set
        and distance function.

    min_samples : int, optional
        The number of samples (or total weight) in a neighborhood for a point
        to be considered as a core point. This includes the point itself.

    metric : string, or callable
        The metric to use when calculating distance between instances in a
        feature array. If metric is a string or callable, it must be one of
        the options allowed by :func:`sklearn.metrics.pairwise_distances` for
        its metric parameter.
        If metric is "precomputed", X is assumed to be a distance matrix and
        must be square. X may be a :term:`Glossary <sparse graph>`, in which
        case only "nonzero" elements may be considered neighbors for DBSCAN.

        .. versionadded:: 0.17
           metric *precomputed* to accept precomputed sparse matrix.

    metric_params : dict, optional
        Additional keyword arguments for the metric function.

        .. versionadded:: 0.19

    algorithm : {'auto', 'ball_tree', 'kd_tree', 'brute', 'daal'}, optional
        The algorithm to be used by the NearestNeighbors module
        to compute pointwise distances and find nearest neighbors.
        See NearestNeighbors module documentation for details.

        If algorithm is set to 'daal', Intel(R) oneAPI Data Analytics Library
        will be used.

    leaf_size : int, optional (default = 30)
        Leaf size passed to BallTree or cKDTree. This can affect the speed
        of the construction and query, as well as the memory required
        to store the tree. The optimal value depends
        on the nature of the problem.

    p : float, optional
        The power of the Minkowski metric to be used to calculate distance
        between points.

    n_jobs : int or None, optional (default=None)
        The number of parallel jobs to run.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors. See :term:`Glossary <n_jobs>`
        for more details.

    Attributes
    ----------
    core_sample_indices_ : array, shape = [n_core_samples]
        Indices of core samples.

    components_ : array, shape = [n_core_samples, n_features]
        Copy of each core sample found by training.

    labels_ : array, shape = [n_samples]
        Cluster labels for each point in the dataset given to fit().
        Noisy samples are given the label -1.

    n_features_in_ : int
        Number of features seen during :term:`fit`.

        .. versionadded:: 0.24

    See Also
    --------
    OPTICS : A similar clustering at multiple values of eps. Our implementation
        is optimized for memory usage.

    Notes
    -----
    For an example, see :ref:`examples/cluster/plot_dbscan.py
    <sphx_glr_auto_examples_cluster_plot_dbscan.py>`.

    This implementation bulk-computes all neighborhood queries, which increases
    the memory complexity to O(n.d) where d is the average number of neighbors,
    while original DBSCAN had memory complexity O(n). It may attract a higher
    memory complexity when querying these nearest neighborhoods, depending
    on the ``algorithm``.

    One way to avoid the query complexity is to pre-compute sparse
    neighborhoods in chunks using
    :func:`NearestNeighbors.radius_neighbors_graph
    <sklearn.neighbors.NearestNeighbors.radius_neighbors_graph>` with
    ``mode='distance'``, then using ``metric='precomputed'`` here.

    Another way to reduce memory and computation time is to remove
    (near-)duplicate points and use ``sample_weight`` instead.

    :class:`cluster.OPTICS` provides a similar clustering with lower memory
    usage.

    References
    ----------
    Ester, M., H. P. Kriegel, J. Sander, and X. Xu, "A Density-Based
    Algorithm for Discovering Clusters in Large Spatial Databases with Noise".
    In: Proceedings of the 2nd International Conference on Knowledge Discovery
    and Data Mining, Portland, OR, AAAI Press, pp. 226-231. 1996

    Schubert, E., Sander, J., Ester, M., Kriegel, H. P., & Xu, X. (2017).
    DBSCAN revisited, revisited: why and how you should (still) use DBSCAN.
    ACM Transactions on Database Systems (TODS), 42(3), 19.

    Examples
    --------
    >>> from sklearn.cluster import DBSCAN
    >>> import numpy as np
    >>> X = np.array([[1, 2], [2, 2], [2, 3],
    ...               [8, 7], [8, 8], [25, 80]])
    >>> clustering = DBSCAN(eps=3, min_samples=2).fit(X)
    >>> clustering.labels_
    array([ 0,  0,  0,  1,  1, -1])
    >>> clustering
    DBSCAN(eps=3, min_samples=2)
    """
    if sklearn_check_version('1.2'):
        _parameter_constraints: dict = {**DBSCAN_original._parameter_constraints}

    def __init__(
        self,
        eps=0.5,
        min_samples=5,
        metric='euclidean',
        metric_params=None,
        algorithm='auto',
        leaf_size=30,
        p=None,
        n_jobs=None,
    ):
        self.eps = eps
        self.min_samples = min_samples
        self.metric = metric
        self.metric_params = metric_params
        self.algorithm = algorithm
        self.leaf_size = leaf_size
        self.p = p
        self.n_jobs = n_jobs

    @support_usm_ndarray()
    def fit(self, X, y=None, sample_weight=None):
        """Perform DBSCAN clustering from features, or distance matrix.

        Parameters
        ----------
        X : array-like or sparse matrix, shape (n_samples, n_features), or \
            (n_samples, n_samples)
            Training instances to cluster, or distances between instances if
            ``metric='precomputed'``. If a sparse matrix is provided, it will
            be converted into a sparse ``csr_matrix``.

        sample_weight : array, shape (n_samples,), optional
            Weight of each sample, such that a sample with a weight of at least
            ``min_samples`` is by itself a core sample; a sample with a
            negative weight may inhibit its eps-neighbor from being core.
            Note that weights are absolute, and default to 1.

        y : Ignored
            Not used, present here for API consistency by convention.

        Returns
        -------
        self : object
            Returns a fitted instance of self.
        """
        if sklearn_check_version("1.2"):
            self._validate_params()
        elif sklearn_check_version("1.1"):
            check_scalar(
                self.eps,
                "eps",
                target_type=numbers.Real,
                min_val=0.0,
                include_boundaries="neither",
            )
            check_scalar(
                self.min_samples,
                "min_samples",
                target_type=numbers.Integral,
                min_val=1,
                include_boundaries="left",
            )
            check_scalar(
                self.leaf_size,
                "leaf_size",
                target_type=numbers.Integral,
                min_val=1,
                include_boundaries="left",
            )
            if self.p is not None:
                check_scalar(
                    self.p,
                    "p",
                    target_type=numbers.Real,
                    min_val=0.0,
                    include_boundaries="left",
                )
            if self.n_jobs is not None:
                check_scalar(self.n_jobs, "n_jobs", target_type=numbers.Integral)
        else:
            if self.eps <= 0.0:
                raise ValueError(f"eps == {self.eps}, must be > 0.0.")

        if sklearn_check_version("1.0"):
            self._check_feature_names(X, reset=True)

        if sample_weight is not None:
            sample_weight = _check_sample_weight(sample_weight, X)

        _patching_status = PatchingConditionsChain(
            "sklearn.cluster.DBSCAN.fit")
        _dal_ready = _patching_status.and_conditions([
            (self.algorithm in ['auto', 'brute'],
                f"'{self.algorithm}' algorithm is not supported. "
                "Only 'auto' and 'brute' algorithms are supported"),
            (self.metric == 'euclidean' or (self.metric == 'minkowski' and self.p == 2),
                f"'{self.metric}' (p={self.p}) metric is not supported. "
                "Only 'euclidean' or 'minkowski' with p=2 metrics are supported."),
            (not sp.issparse(X), "X is sparse. Sparse input is not supported.")
        ])

        _patching_status.write_log()
        if _dal_ready:
            X = check_array(X, accept_sparse='csr', dtype=[np.float64, np.float32])
            core_ind, assignments = _daal_dbscan(
                X,
                self.eps,
                self.min_samples,
                sample_weight=sample_weight
            )
            self.core_sample_indices_ = core_ind
            self.labels_ = assignments
            self.components_ = np.take(X, core_ind, axis=0)
            self.n_features_in_ = X.shape[1]
            return self
        return super().fit(X, y, sample_weight=sample_weight)

    @support_usm_ndarray()
    def fit_predict(self, X, y=None, sample_weight=None):
        """
        Compute clusters from a data or distance matrix and predict labels.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features), or \
            (n_samples, n_samples)
            Training instances to cluster, or distances between instances if
            ``metric='precomputed'``. If a sparse matrix is provided, it will
            be converted into a sparse ``csr_matrix``.

        y : Ignored
            Not used, present here for API consistency by convention.

        sample_weight : array-like of shape (n_samples,), default=None
            Weight of each sample, such that a sample with a weight of at least
            ``min_samples`` is by itself a core sample; a sample with a
            negative weight may inhibit its eps-neighbor from being core.
            Note that weights are absolute, and default to 1.

        Returns
        -------
        labels : ndarray of shape (n_samples,)
            Cluster labels. Noisy samples are given the label -1.
        """
        return super().fit_predict(X, y, sample_weight)



# ===============================================================================
# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============================================================================

import platform
from daal4py.sklearn._utils import daal_check_version

if "Windows" in platform.system():
    import os
    import sys
    import site
    path_to_env = site.getsitepackages()[0]
    path_to_libs = os.path.join(path_to_env, "Library", "bin")
    if sys.version_info.minor >= 8:
        if 'DALROOT' in os.environ:
            dal_root_redist = os.path.join(
                os.environ['DALROOT'], "redist", "intel64")
            if os.path.exists(dal_root_redist):
                os.add_dll_directory(dal_root_redist)
        os.add_dll_directory(path_to_libs)
    os.environ['PATH'] = path_to_libs + os.pathsep + os.environ['PATH']

try:
    import onedal._onedal_py_dpc as _backend
    _is_dpc_backend = True
except ImportError:
    import onedal._onedal_py_host as _backend
    _is_dpc_backend = False


__all__ = ['decomposition', 'ensemble', 'neighbors', 'primitives', 'svm']

if _is_dpc_backend:
    __all__.append('spmd')

if daal_check_version((2023, 'P', 100)):
    __all__ += ['basic_statistics', 'linear_model']

    if _is_dpc_backend:
        __all__ += ['spmd.basic_statistics', 'spmd.decomposition',
                    'spmd.linear_model', 'spmd.neighbors']



#===============================================================================
# Copyright 2020 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

import numpy as np
import numbers
import daal4py
from scipy import sparse as sp
from sklearn.utils import check_array, check_X_y
from sklearn.linear_model._coordinate_descent import ElasticNet as ElasticNet_original
from sklearn.linear_model._coordinate_descent import Lasso as Lasso_original
from daal4py.sklearn._utils import (
    make2d, getFPType, get_patch_message, sklearn_check_version, PatchingConditionsChain)
if sklearn_check_version('1.0') and not sklearn_check_version('1.2'):
    from sklearn.linear_model._base import _deprecate_normalize
if sklearn_check_version('1.1') and not sklearn_check_version('1.2'):
    from sklearn.utils import check_scalar

import logging

# only for compliance with Sklearn
import warnings
from sklearn.exceptions import ConvergenceWarning
from sklearn.preprocessing import normalize

from .._device_offload import support_usm_ndarray


def _daal4py_check(self, X, y, check_input):
    _fptype = getFPType(X)

    # check alpha
    if self.alpha == 0:
        warnings.warn("With alpha=0, this algorithm does not converge "
                      "well. You are advised to use the LinearRegression "
                      "estimator", stacklevel=2)

    # check l1_ratio
    if not isinstance(self.l1_ratio, numbers.Number) or \
            self.l1_ratio < 0 or self.l1_ratio > 1:
        raise ValueError("l1_ratio must be between 0 and 1; "
                         f"got l1_ratio={self.l1_ratio}")

    # check precompute
    if isinstance(self.precompute, np.ndarray):
        if check_input:
            check_array(self.precompute, dtype=_fptype)
        self.precompute = make2d(self.precompute)
    else:
        if self.precompute not in [False, True, 'auto']:
            raise ValueError("precompute should be one of True, False, "
                             "'auto' or array-like. Got %r" % self.precompute)

    # check selection
    if self.selection not in ['random', 'cyclic']:
        raise ValueError("selection should be either random or cyclic.")


def _daal4py_fit_enet(self, X, y_, check_input):

    # appropriate checks
    _daal4py_check(self, X, y_, check_input)
    X = make2d(X)
    y = make2d(y_)
    _fptype = getFPType(X)

    # only for dual_gap computation, it is not required for Intel(R) oneAPI
    # Data Analytics Library
    self._X = X
    if sklearn_check_version('0.23'):
        self.n_features_in_ = X.shape[1]
    self._y = y

    penalty_L1 = np.asarray(self.alpha * self.l1_ratio, dtype=X.dtype)
    penalty_L2 = np.asarray(self.alpha * (1.0 - self.l1_ratio), dtype=X.dtype)
    if (penalty_L1.size != 1 or penalty_L2.size != 1):
        raise ValueError("alpha or l1_ratio length is wrong")
    penalty_L1 = penalty_L1.reshape((1, -1))
    penalty_L2 = penalty_L2.reshape((1, -1))

    #normalizing and centering
    X_offset = np.zeros(X.shape[1], dtype=X.dtype)
    X_scale = np.ones(X.shape[1], dtype=X.dtype)
    if y.ndim == 1:
        y_offset = X.dtype.type(0)
    else:
        y_offset = np.zeros(y.shape[1], dtype=X.dtype)

    if sklearn_check_version('1.2'):
        _normalize = False
    else:
        _normalize = self._normalize if sklearn_check_version('1.0') else self.normalize
    if self.fit_intercept:
        X_offset = np.average(X, axis=0)
        if _normalize:
            if self.copy_X:
                X = np.copy(X) - X_offset
            else:
                X -= X_offset
            X, X_scale = normalize(X, axis=0, copy=False, return_norm=True)
            y_offset = np.average(y, axis=0)
            y = y - y_offset

    # only for compliance with Sklearn
    if isinstance(self.precompute, np.ndarray) and self.fit_intercept and \
       not np.allclose(X_offset, np.zeros(X.shape[1])) or \
       _normalize and not np.allclose(X_scale, np.ones(X.shape[1])):
        warnings.warn("Gram matrix was provided but X was centered"
                      " to fit intercept, "
                      "or X was normalized : recomputing Gram matrix.",
                      UserWarning)

    mse_alg = daal4py.optimization_solver_mse(
        numberOfTerms=X.shape[0],
        fptype=_fptype,
        method='defaultDense'
    )
    mse_alg.setup(X, y, None)

    cd_solver = daal4py.optimization_solver_coordinate_descent(
        function=mse_alg,
        fptype=_fptype,
        method='defaultDense',
        selection=self.selection,
        seed=0 if self.random_state is None else self.random_state,
        nIterations=self.max_iter,
        positive=self.positive,
        accuracyThreshold=self.tol,
    )

    # set warm_start
    if self.warm_start and hasattr(self, "coef_") and \
            isinstance(self.coef_, np.ndarray):
        n_rows = y.shape[1]
        n_cols = X.shape[1] + 1
        inputArgument = np.zeros((n_rows, n_cols), dtype=_fptype)
        for i in range(n_rows):
            inputArgument[i][0] = self.intercept_ if (
                n_rows == 1) else self.intercept_[i]
            inputArgument[i][1:] = self.coef_[:].copy(order='C') if (
                n_rows == 1) else self.coef_[i, :].copy(order='C')
        cd_solver.setup(inputArgument)
    doUse_condition = self.copy_X is False or \
        (self.fit_intercept and _normalize and self.copy_X)
    elastic_net_alg = daal4py.elastic_net_training(
        fptype=_fptype,
        method='defaultDense',
        interceptFlag=(
            self.fit_intercept is True),
        dataUseInComputation='doUse' if doUse_condition else 'doNotUse',
        penaltyL1=penalty_L1,
        penaltyL2=penalty_L2,
        optimizationSolver=cd_solver
    )
    try:
        if isinstance(self.precompute, np.ndarray):
            elastic_net_res = elastic_net_alg.compute(
                data=X, dependentVariables=y, gramMatrix=self.precompute)
        else:
            elastic_net_res = elastic_net_alg.compute(
                data=X, dependentVariables=y)
    except RuntimeError:
        return None

    # set coef_ and intersept_ results
    elastic_net_model = elastic_net_res.model
    self.daal_model_ = elastic_net_model

    # update coefficients if normalizing and centering
    if self.fit_intercept and _normalize:
        elastic_net_model.Beta[:, 1:] = elastic_net_model.Beta[:, 1:] / X_scale
        elastic_net_model.Beta[:, 0] = (
            y_offset - np.dot(X_offset, elastic_net_model.Beta[:, 1:].T)).T

    coefs = elastic_net_model.Beta

    self.intercept_ = coefs[:, 0].copy(order='C')
    self.coef_ = coefs[:, 1:].copy(order='C')

    # only for compliance with Sklearn
    if y.shape[1] == 1:
        self.coef_ = np.ravel(self.coef_)
    self.intercept_ = np.ravel(self.intercept_)
    if self.intercept_.shape[0] == 1:
        self.intercept_ = self.intercept_[0]

    # set n_iter_
    n_iter = cd_solver.__get_result__().nIterations[0][0]
    if y.shape[1] == 1:
        self.n_iter_ = n_iter
    else:
        self.n_iter_ = np.full(y.shape[1], n_iter)

    # only for compliance with Sklearn
    if self.max_iter == n_iter + 1:
        warnings.warn("Objective did not converge. You might want to "
                      "increase the number of iterations.", ConvergenceWarning)

    return self


def _daal4py_predict_enet(self, X):
    X = make2d(X)
    _fptype = getFPType(self.coef_)

    elastic_net_palg = daal4py.elastic_net_prediction(
        fptype=_fptype,
        method='defaultDense'
    )
    if sklearn_check_version('0.23'):
        if self.n_features_in_ != X.shape[1]:
            raise ValueError(f'X has {X.shape[1]} features, '
                             f'but ElasticNet is expecting '
                             f'{self.n_features_in_} features as input')
    elastic_net_res = elastic_net_palg.compute(X, self.daal_model_)

    res = elastic_net_res.prediction

    if res.shape[1] == 1 and self.coef_.ndim == 1:
        res = np.ravel(res)
    return res


def _daal4py_fit_lasso(self, X, y_, check_input):

    # appropriate checks
    _daal4py_check(self, X, y_, check_input)
    X = make2d(X)
    y = make2d(y_)
    _fptype = getFPType(X)

    # only for dual_gap computation, it is not required for Intel(R) oneAPI
    # Data Analytics Library
    self._X = X
    if sklearn_check_version('0.23'):
        self.n_features_in_ = X.shape[1]
    self._y = y

    #normalizing and centering
    X_offset = np.zeros(X.shape[1], dtype=X.dtype)
    X_scale = np.ones(X.shape[1], dtype=X.dtype)
    if y.ndim == 1:
        y_offset = X.dtype.type(0)
    else:
        y_offset = np.zeros(y.shape[1], dtype=X.dtype)

    if sklearn_check_version('1.2'):
        _normalize = False
    else:
        _normalize = self._normalize if sklearn_check_version('1.0') else self.normalize
    if self.fit_intercept:
        X_offset = np.average(X, axis=0)
        if _normalize:
            if self.copy_X:
                X = np.copy(X) - X_offset
            else:
                X -= X_offset
            X, X_scale = normalize(X, axis=0, copy=False, return_norm=True)
            y_offset = np.average(y, axis=0)
            y = y - y_offset

    # only for compliance with Sklearn
    if isinstance(self.precompute, np.ndarray) and \
       self.fit_intercept and not np.allclose(
            X_offset, np.zeros(X.shape[1])) or \
       _normalize and not np.allclose(X_scale, np.ones(X.shape[1])):
        warnings.warn("Gram matrix was provided but X was centered"
                      " to fit intercept, "
                      "or X was normalized : recomputing Gram matrix.",
                      UserWarning)

    mse_alg = daal4py.optimization_solver_mse(
        numberOfTerms=X.shape[0],
        fptype=_fptype,
        method='defaultDense'
    )
    mse_alg.setup(X, y, None)

    cd_solver = daal4py.optimization_solver_coordinate_descent(
        function=mse_alg,
        fptype=_fptype,
        method='defaultDense',
        selection=self.selection,
        seed=0 if self.random_state is None else self.random_state,
        nIterations=self.max_iter,
        positive=self.positive,
        accuracyThreshold=self.tol
    )

    # set warm_start
    if self.warm_start and hasattr(self, "coef_") and \
            isinstance(self.coef_, np.ndarray):
        n_rows = y.shape[1]
        n_cols = X.shape[1] + 1
        inputArgument = np.zeros((n_rows, n_cols), dtype=_fptype)
        for i in range(n_rows):
            inputArgument[i][0] = self.intercept_ if (
                n_rows == 1) else self.intercept_[i]
            inputArgument[i][1:] = self.coef_[:].copy(order='C') if (
                n_rows == 1) else self.coef_[i, :].copy(order='C')
        cd_solver.setup(inputArgument)
    doUse_condition = self.copy_X is False or \
        (self.fit_intercept and _normalize and self.copy_X)
    lasso_alg = daal4py.lasso_regression_training(
        fptype=_fptype,
        method='defaultDense',
        interceptFlag=(self.fit_intercept is True),
        dataUseInComputation='doUse' if doUse_condition else 'doNotUse',
        lassoParameters=np.asarray(
            self.alpha, dtype=X.dtype
        ).reshape((1, -1)),
        optimizationSolver=cd_solver,
    )
    try:
        if isinstance(self.precompute, np.ndarray):
            lasso_res = lasso_alg.compute(
                data=X, dependentVariables=y, gramMatrix=self.precompute)
        else:
            lasso_res = lasso_alg.compute(data=X, dependentVariables=y)
    except RuntimeError:
        return None

    # set coef_ and intersept_ results
    lasso_model = lasso_res.model
    self.daal_model_ = lasso_model

    # update coefficients if normalizing and centering
    if self.fit_intercept and _normalize:
        lasso_model.Beta[:, 1:] = lasso_model.Beta[:, 1:] / X_scale
        lasso_model.Beta[:, 0] = \
            (y_offset - np.dot(X_offset, lasso_model.Beta[:, 1:].T)).T

    coefs = lasso_model.Beta

    self.intercept_ = coefs[:, 0].copy(order='C')
    self.coef_ = coefs[:, 1:].copy(order='C')

    # only for compliance with Sklearn
    if y.shape[1] == 1:
        self.coef_ = np.ravel(self.coef_)
    self.intercept_ = np.ravel(self.intercept_)
    if self.intercept_.shape[0] == 1:
        self.intercept_ = self.intercept_[0]

    # set n_iter_
    n_iter = cd_solver.__get_result__().nIterations[0][0]
    if y.shape[1] == 1:
        self.n_iter_ = n_iter
    else:
        self.n_iter_ = np.full(y.shape[1], n_iter)

    # only for compliance with Sklearn
    if (self.max_iter == n_iter + 1):
        warnings.warn("Objective did not converge. You might want to "
                      "increase the number of iterations.", ConvergenceWarning)

    return self


def _daal4py_predict_lasso(self, X):
    X = make2d(X)
    _fptype = getFPType(self.coef_)

    lasso_palg = daal4py.lasso_regression_prediction(
        fptype=_fptype,
        method='defaultDense'
    )
    if sklearn_check_version('0.23'):
        if self.n_features_in_ != X.shape[1]:
            raise ValueError(f'X has {X.shape[1]} features, '
                             f'but Lasso is expecting '
                             f'{self.n_features_in_} features as input')
    lasso_res = lasso_palg.compute(X, self.daal_model_)

    res = lasso_res.prediction

    if res.shape[1] == 1 and self.coef_.ndim == 1:
        res = np.ravel(res)
    return res


def _fit(self, X, y, sample_weight=None, check_input=True):
    if sklearn_check_version('1.0'):
        self._check_feature_names(X, reset=True)
    if sklearn_check_version("1.2"):
        self._validate_params()
    elif sklearn_check_version('1.1'):
        check_scalar(
            self.alpha,
            "alpha",
            target_type=numbers.Real,
            min_val=0.0,
        )
        if self.alpha == 0:
            warnings.warn(
                "With alpha=0, this algorithm does not converge "
                "well. You are advised to use the LinearRegression "
                "estimator",
                stacklevel=2,
            )
        if isinstance(self.precompute, str):
            raise ValueError(
                "precompute should be one of True, False or array-like. Got %r"
                % self.precompute
            )
        check_scalar(
            self.l1_ratio,
            "l1_ratio",
            target_type=numbers.Real,
            min_val=0.0,
            max_val=1.0,
        )
        if self.max_iter is not None:
            check_scalar(
                self.max_iter, "max_iter", target_type=numbers.Integral, min_val=1
            )
        check_scalar(self.tol, "tol", target_type=numbers.Real, min_val=0.0)
    # check X and y
    if check_input:
        X, y = check_X_y(
            X,
            y,
            copy=False,
            accept_sparse='csc',
            dtype=[np.float64, np.float32],
            multi_output=True,
            y_numeric=True,
        )
        y = check_array(y, copy=False, dtype=X.dtype.type, ensure_2d=False)

    if not sp.issparse(X):
        self.fit_shape_good_for_daal_ = \
            True if X.ndim <= 1 else True if X.shape[0] >= X.shape[1] else False
    else:
        self.fit_shape_good_for_daal_ = False

    class_name = self.__class__.__name__
    class_inst = ElasticNet if class_name == 'ElasticNet' else Lasso

    _function_name = f"sklearn.linear_model.{class_name}.fit"
    _patching_status = PatchingConditionsChain(
        _function_name)
    _dal_ready = _patching_status.and_conditions([
        (not sp.issparse(X), "X is sparse. Sparse input is not supported."),
        (self.fit_shape_good_for_daal_,
            "The shape of X does not satisfy oneDAL requirements: "
            "number of features > number of samples."),
        (X.dtype == np.float64 or X.dtype == np.float32,
            f"'{X.dtype}' X data type is not supported. "
            "Only np.float32 and np.float64 are supported."),
        (sample_weight is None, "Sample weights are not supported.")])
    _patching_status.write_log()

    if not _dal_ready:
        if hasattr(self, 'daal_model_'):
            del self.daal_model_
        if sklearn_check_version('0.23'):
            res_new = super(class_inst, self).fit(
                X, y, sample_weight=sample_weight, check_input=check_input)
        else:
            res_new = super(class_inst, self).fit(
                X, y, check_input=check_input)
        self._gap = res_new.dual_gap_
        return res_new
    self.n_iter_ = None
    self._gap = None

    if not check_input:
        # only for compliance with Sklearn,
        # this assert is not required for Intel(R) oneAPI Data
        # Analytics Library
        print(type(X), X.flags['F_CONTIGUOUS'])
        if isinstance(X, np.ndarray) and \
                X.flags['F_CONTIGUOUS'] is False:
            # print(X.flags)
            raise ValueError("ndarray is not Fortran contiguous")

    if sklearn_check_version('1.0') and not sklearn_check_version('1.2'):
        self._normalize = _deprecate_normalize(
            self.normalize,
            default=False,
            estimator_name=class_name
        )

    # only for pass tests
    # "check_estimators_fit_returns_self(readonly_memmap=True) and
    # check_regressors_train(readonly_memmap=True)
    if not X.flags.writeable:
        X = np.copy(X)
    if not y.flags.writeable:
        y = np.copy(y)

    if class_name == "ElasticNet":
        res = _daal4py_fit_enet(self, X, y, check_input=check_input)
    else:
        res = _daal4py_fit_lasso(self, X, y, check_input=check_input)
    if res is None:
        if hasattr(self, 'daal_model_'):
            del self.daal_model_
        logging.info(
            _function_name + ": " + get_patch_message("sklearn_after_daal")
        )
        if sklearn_check_version('0.23'):
            res_new = super(class_inst, self).fit(
                X, y, sample_weight=sample_weight, check_input=check_input)
        else:
            res_new = super(class_inst, self).fit(
                X, y, check_input=check_input)
        self._gap = res_new.dual_gap_
        return res_new
    return res


def _dual_gap(self):
    if (self._gap is None):
        l1_reg = self.alpha * self.l1_ratio * self._X.shape[0]
        l2_reg = self.alpha * (1.0 - self.l1_ratio) * self._X.shape[0]
        n_targets = self._y.shape[1]

        if (n_targets == 1):
            self._gap = self.tol + 1.0
            X_offset = np.average(self._X, axis=0)
            y_offset = np.average(self._y, axis=0)
            coef = np.reshape(self.coef_, (self.coef_.shape[0], 1))
            R = (self._y - y_offset) - np.dot((self._X - X_offset), coef)
            XtA = np.dot((self._X - X_offset).T, R) - l2_reg * coef
            R_norm2 = np.dot(R.T, R)
            coef_norm2 = np.dot(self.coef_, self.coef_)
            dual_norm_XtA = np.max(
                XtA) if self.positive else np.max(np.abs(XtA))
            if dual_norm_XtA > l1_reg:
                const = l1_reg / dual_norm_XtA
                A_norm2 = R_norm2 * (const ** 2)
                self._gap = 0.5 * (R_norm2 + A_norm2)
            else:
                const = 1.0
                self._gap = R_norm2
            l1_norm = np.sum(np.abs(self.coef_))
            tmp = l1_reg * l1_norm
            tmp -= const * np.dot(R.T, (self._y - y_offset))
            tmp += 0.5 * l2_reg * (1 + const ** 2) * coef_norm2
            self._gap += tmp
            self._gap = self._gap[0][0]
        else:
            self._gap = np.full(n_targets, self.tol + 1.0)
            X_offset = np.average(self._X, axis=0)
            y_offset = np.average(self._y, axis=0)
            for k in range(n_targets):
                R = (self._y[:, k] - y_offset[k]) - \
                    np.dot((self._X - X_offset), self.coef_[k, :].T)
                XtA = np.dot((self._X - X_offset).T, R) - \
                    l2_reg * self.coef_[k, :].T
                R_norm2 = np.dot(R.T, R)
                coef_norm2 = np.dot(self.coef_[k, :], self.coef_[k, :].T)
                dual_norm_XtA = np.max(
                    XtA) if self.positive else np.max(np.abs(XtA))
                if dual_norm_XtA > l1_reg:
                    const = l1_reg / dual_norm_XtA
                    A_norm2 = R_norm2 * (const ** 2)
                    self._gap[k] = 0.5 * (R_norm2 + A_norm2)
                else:
                    const = 1.0
                    self._gap[k] = R_norm2
                l1_norm = np.sum(np.abs(self.coef_[k, :]))
                tmp = l1_reg * l1_norm
                tmp -= const * np.dot(R.T, (self._y[:, k] - y_offset[k]))
                tmp += 0.5 * l2_reg * (1 + const ** 2) * coef_norm2
                self._gap[k] += tmp
    return self._gap


class ElasticNet(ElasticNet_original):
    __doc__ = ElasticNet_original.__doc__

    if sklearn_check_version('1.2'):
        _parameter_constraints: dict = {**ElasticNet_original._parameter_constraints}

        def __init__(
            self,
            alpha=1.0,
            l1_ratio=0.5,
            fit_intercept=True,
            precompute=False,
            max_iter=1000,
            copy_X=True,
            tol=1e-4,
            warm_start=False,
            positive=False,
            random_state=None,
            selection='cyclic',
        ):
            super(ElasticNet, self).__init__(
                alpha=alpha,
                l1_ratio=l1_ratio,
                fit_intercept=fit_intercept,
                precompute=precompute,
                max_iter=max_iter,
                copy_X=copy_X,
                tol=tol,
                warm_start=warm_start,
                positive=positive,
                random_state=random_state,
                selection=selection,
            )
    else:
        def __init__(
            self,
            alpha=1.0,
            l1_ratio=0.5,
            fit_intercept=True,
            normalize="deprecated" if sklearn_check_version('1.0') else False,
            precompute=False,
            max_iter=1000,
            copy_X=True,
            tol=1e-4,
            warm_start=False,
            positive=False,
            random_state=None,
            selection='cyclic',
        ):
            super(ElasticNet, self).__init__(
                alpha=alpha,
                l1_ratio=l1_ratio,
                fit_intercept=fit_intercept,
                normalize=normalize,
                precompute=precompute,
                max_iter=max_iter,
                copy_X=copy_X,
                tol=tol,
                warm_start=warm_start,
                positive=positive,
                random_state=random_state,
                selection=selection,
            )

    if sklearn_check_version('0.23'):
        @support_usm_ndarray()
        def fit(self, X, y, sample_weight=None, check_input=True):
            """
            Fit model with coordinate descent.

            Parameters
            ----------
            X : {ndarray, sparse matrix} of (n_samples, n_features)
                Data.

            y : {ndarray, sparse matrix} of shape (n_samples,) or \
                (n_samples, n_targets)
                Target. Will be cast to X's dtype if necessary.

            sample_weight : float or array-like of shape (n_samples,), default=None
                Sample weights. Internally, the `sample_weight` vector will be
                rescaled to sum to `n_samples`.

                .. versionadded:: 0.23

            check_input : bool, default=True
                Allow to bypass several input checking.
                Don't use this parameter unless you know what you do.

            Returns
            -------
            self : object
                Fitted estimator.

            Notes
            -----
            Coordinate descent is an algorithm that considers each column of
            data at a time hence it will automatically convert the X input
            as a Fortran-contiguous numpy array if necessary.

            To avoid memory re-allocation it is advised to allocate the
            initial data in memory directly using that format.
            """
            return _fit(self, X, y, sample_weight=sample_weight, check_input=check_input)
    else:
        @support_usm_ndarray()
        def fit(self, X, y, check_input=True):
            """
            Fit model with coordinate descent.

            Parameters
            ----------
            X : ndarray or scipy.sparse matrix, (n_samples, n_features)
                Data

            y : ndarray, shape (n_samples,) or (n_samples, n_targets)
                Target. Will be cast to X's dtype if necessary

            check_input : boolean, (default=True)
                Allow to bypass several input checking.
                Don't use this parameter unless you know what you do.

            Notes
            -----

            Coordinate descent is an algorithm that considers each column of
            data at a time hence it will automatically convert the X input
            as a Fortran-contiguous numpy array if necessary.

            To avoid memory re-allocation it is advised to allocate the
            initial data in memory directly using that format.
            """
            return _fit(self, X, y, check_input=check_input)

    @support_usm_ndarray()
    def predict(self, X):
        """Predict using the linear model

        Parameters
        ----------
        X : array-like or sparse matrix, shape = (n_samples, n_features)
            Samples.

        Returns
        -------
        C : array, shape = (n_samples,)
            Returns predicted values.
        """

        if sklearn_check_version('1.0'):
            self._check_feature_names(X, reset=False)

        X = check_array(
            X,
            accept_sparse=['csr', 'csc', 'coo'],
            dtype=[np.float64, np.float32]
        )
        good_shape_for_daal = \
            True if X.ndim <= 1 else True if X.shape[0] >= X.shape[1] else False

        _patching_status = PatchingConditionsChain(
            "sklearn.linear_model.ElasticNet.predict")
        _dal_ready = _patching_status.and_conditions([
            (hasattr(self, 'daal_model_'), "oneDAL model was not trained."),
            (not sp.issparse(X), "X is sparse. Sparse input is not supported."),
            (good_shape_for_daal,
                "The shape of X does not satisfy oneDAL requirements: "
                "number of features > number of samples.")])
        _patching_status.write_log()

        if not _dal_ready:
            return self._decision_function(X)
        return _daal4py_predict_enet(self, X)

    @property
    def dual_gap_(self):
        return _dual_gap(self)

    @dual_gap_.setter
    def dual_gap_(self, value):
        self._gap = value

    @dual_gap_.deleter
    def dual_gap_(self):
        self._gap = None


class Lasso(Lasso_original):
    __doc__ = Lasso_original.__doc__

    if sklearn_check_version('1.2'):
        def __init__(
            self,
            alpha=1.0,
            fit_intercept=True,
            precompute=False,
            copy_X=True,
            max_iter=1000,
            tol=1e-4,
            warm_start=False,
            positive=False,
            random_state=None,
            selection='cyclic',
        ):
            self.l1_ratio = 1.0
            super().__init__(
                alpha=alpha,
                fit_intercept=fit_intercept,
                precompute=precompute,
                copy_X=copy_X,
                max_iter=max_iter,
                tol=tol,
                warm_start=warm_start,
                positive=positive,
                random_state=random_state,
                selection=selection,
            )
    else:
        def __init__(
            self,
            alpha=1.0,
            fit_intercept=True,
            normalize="deprecated" if sklearn_check_version('1.0') else False,
            precompute=False,
            copy_X=True,
            max_iter=1000,
            tol=1e-4,
            warm_start=False,
            positive=False,
            random_state=None,
            selection='cyclic',
        ):
            self.l1_ratio = 1.0
            super().__init__(
                alpha=alpha,
                fit_intercept=fit_intercept,
                normalize=normalize,
                precompute=precompute,
                copy_X=copy_X,
                max_iter=max_iter,
                tol=tol,
                warm_start=warm_start,
                positive=positive,
                random_state=random_state,
                selection=selection,
            )

    if sklearn_check_version('0.23'):
        @support_usm_ndarray()
        def fit(self, X, y, sample_weight=None, check_input=True):
            """
            Fit model with coordinate descent.

            Parameters
            ----------
            X : {ndarray, sparse matrix} of (n_samples, n_features)
                Data.

            y : {ndarray, sparse matrix} of shape (n_samples,) or \
                (n_samples, n_targets)
                Target. Will be cast to X's dtype if necessary.

            sample_weight : float or array-like of shape (n_samples,), default=None
                Sample weights. Internally, the `sample_weight` vector will be
                rescaled to sum to `n_samples`.

                .. versionadded:: 0.23

            check_input : bool, default=True
                Allow to bypass several input checking.
                Don't use this parameter unless you know what you do.

            Returns
            -------
            self : object
                Fitted estimator.

            Notes
            -----
            Coordinate descent is an algorithm that considers each column of
            data at a time hence it will automatically convert the X input
            as a Fortran-contiguous numpy array if necessary.

            To avoid memory re-allocation it is advised to allocate the
            initial data in memory directly using that format.
            """
            return _fit(self, X, y, sample_weight, check_input)
    else:
        @support_usm_ndarray()
        def fit(self, X, y, check_input=True):
            """
            Fit model with coordinate descent.

            Parameters
            ----------
            X : ndarray or scipy.sparse matrix, (n_samples, n_features)
                Data

            y : ndarray, shape (n_samples,) or (n_samples, n_targets)
                Target. Will be cast to X's dtype if necessary

            check_input : boolean, (default=True)
                Allow to bypass several input checking.
                Don't use this parameter unless you know what you do.

            Notes
            -----

            Coordinate descent is an algorithm that considers each column of
            data at a time hence it will automatically convert the X input
            as a Fortran-contiguous numpy array if necessary.

            To avoid memory re-allocation it is advised to allocate the
            initial data in memory directly using that format.
            """
            return _fit(self, X, y, check_input)

    @support_usm_ndarray()
    def predict(self, X):
        """Predict using the linear model

        Parameters
        ----------
        X : array-like or sparse matrix, shape = (n_samples, n_features)
            Samples.

        Returns
        -------
        C : array, shape = (n_samples,)
            Returns predicted values.
        """
        if sklearn_check_version('1.0'):
            self._check_feature_names(X, reset=False)
        X = check_array(
            X,
            accept_sparse=['csr', 'csc', 'coo'],
            dtype=[np.float64, np.float32]
        )
        good_shape_for_daal = \
            True if X.ndim <= 1 else True if X.shape[0] >= X.shape[1] else False

        _patching_status = PatchingConditionsChain(
            "sklearn.linear_model.Lasso.predict")
        _dal_ready = _patching_status.and_conditions([
            (hasattr(self, 'daal_model_'), "oneDAL model was not trained."),
            (not sp.issparse(X), "X is sparse. Sparse input is not supported."),
            (good_shape_for_daal,
                "The shape of X does not satisfy oneDAL requirements: "
                "number of features > number of samples.")])
        _patching_status.write_log()

        if not _dal_ready:
            return self._decision_function(X)
        return _daal4py_predict_lasso(self, X)

    @property
    def dual_gap_(self):
        return _dual_gap(self)

    @dual_gap_.setter
    def dual_gap_(self, value):
        self._gap = value

    @dual_gap_.deleter
    def dual_gap_(self):
        self._gap = None



#===============================================================================
# Copyright 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

import os
import subprocess
import unittest
from daal4py.sklearn._utils import get_daal_version
test_path = os.path.abspath(os.path.dirname(__file__))
unittest_data_path = os.path.join(test_path, "unittest_data")
examples_path = os.path.join(
    os.path.dirname(test_path), "examples", "sklearnex")

print('Testing examples_sklearnex')
# First item is major version - 2021,
# second is minor+patch - 0110,
# third item is status - B
sklearnex_version = get_daal_version()
print('oneDAL version:', sklearnex_version)


class TestsklearnexExamples(unittest.TestCase):
    '''Class for testing sklernex examples'''
    # Get a list of all Python files in the examples directory
    pass


def test_generator(file):
    def testit(self):
        # Run the script and capture its exit code
        process = subprocess.run(
            ['python', os.path.join(examples_path, file)],
            stdout=subprocess.PIPE, stderr=subprocess.PIPE,
            check=True)  # nosec
        exit_code = process.returncode

        # Assert that the exit code is 0
        self.assertEqual(exit_code, 0)

    setattr(TestsklearnexExamples, 'test_' + os.path.splitext(file)[0], testit)
    print("Generating tests for " + os.path.splitext(file)[0])


files = [f for f in os.listdir(examples_path) if f.endswith(".py") and 'spmd' not in f]

for file in files:
    test_generator(file)

if __name__ == '__main__':
    unittest.main()



# ===============================================================================
# Copyright 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============================================================================

from sklearn.base import BaseEstimator
from abc import ABCMeta, abstractmethod

import numpy as np
from numbers import Number

from ..common._policy import _get_policy

from ..datatypes._data_conversion import (
    from_table,
    to_table,
    _convert_to_supported,
    _convert_to_dataframe)
from onedal import _backend


class BaseBasicStatistics(metaclass=ABCMeta):
    @abstractmethod
    def __init__(self, result_options, algorithm):
        self.options = result_options
        self.algorithm = algorithm

    @staticmethod
    def get_all_result_options():
        return ["min", "max", "sum", "mean",
                "variance", "variation", "sum_squares",
                "standard_deviation", "sum_squares_centered",
                "second_order_raw_moment"]

    def _get_policy(self, queue, *data):
        return _get_policy(queue, *data)

    def _get_result_options(self, options):
        if options == "all":
            options = self.get_all_result_options()
        if isinstance(options, list):
            options = "|".join(options)
        assert isinstance(options, str)
        return options

    def _get_onedal_params(self, dtype=np.float32):
        options = self._get_result_options(self.options)
        return {
            'fptype': 'float' if dtype == np.float32 else 'double',
            'method': self.algorithm, 'result_option': options,
        }

    def _compute(self, data, weights, module, queue):
        policy = self._get_policy(queue, data, weights)

        data_loc, weights_loc = _convert_to_dataframe(policy, data, weights)

        data_loc, weights_loc = _convert_to_supported(
            policy, data_loc, weights_loc)

        params = self._get_onedal_params(data_loc.dtype)
        data_table, weights_table = to_table(data_loc, weights_loc)

        result = module.train(policy, params, data_table, weights_table)

        options = self._get_result_options(self.options)
        options = options.split("|")

        res = {opt: getattr(result, opt) for opt in options}

        return {k: from_table(v).ravel() for k, v in res.items()}


class BasicStatistics(BaseBasicStatistics):
    """
    Basic Statistics oneDAL implementation.
    """

    def __init__(
            self,
            result_options="all",
            *,
            algorithm="by_default",
            **kwargs):
        super().__init__(result_options, algorithm)

    def compute(self, data, weights=None, queue=None):
        return super()._compute(data, weights, _backend.basic_statistics.compute, queue)



#===============================================================================
# Copyright 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

from abc import ABC

from ...common._spmd_policy import _get_spmd_policy

from onedal.ensemble import RandomForestClassifier as RandomForestClassifier_Batch
from onedal.ensemble import RandomForestRegressor as RandomForestRegressor_Batch


class BaseForestSPMD(ABC):
    def _get_policy(self, queue, *data):
        return _get_spmd_policy(queue)


class RandomForestClassifier(BaseForestSPMD, RandomForestClassifier_Batch):
    pass


class RandomForestRegressor(BaseForestSPMD, RandomForestRegressor_Batch):
    pass



# ===============================================================================
# Copyright 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============================================================================

from sklearn.base import BaseEstimator
from abc import ABCMeta, abstractmethod

import numpy as np
from numbers import Number

from daal4py.sklearn._utils import (get_dtype, make2d)
from ..datatypes import (
    _check_X_y,
    _num_features,
    _check_array,
    _get_2d_shape,
    _check_n_features,
    _convert_to_supported)

from ..common._mixin import RegressorMixin
from ..common._policy import _get_policy
from ..common._estimator_checks import _check_is_fitted
from ..datatypes._data_conversion import from_table, to_table
from onedal import _backend


class BaseLinearRegression(BaseEstimator, metaclass=ABCMeta):
    @abstractmethod
    def __init__(self, fit_intercept, copy_X, algorithm):
        self.fit_intercept = fit_intercept
        self.algorithm = algorithm
        self.copy_X = copy_X

    def _get_policy(self, queue, *data):
        return _get_policy(queue, *data)

    def _get_onedal_params(self, dtype=np.float32):
        intercept = 'intercept|' if self.fit_intercept else ''
        return {
            'fptype': 'float' if dtype == np.float32 else 'double',
            'method': self.algorithm, 'intercept': self.fit_intercept,
            'result_option': (intercept + 'coefficients'),
        }

    def _fit(self, X, y, module, queue):
        policy = self._get_policy(queue, X, y)

        X_loc, y_loc = X, y
        if not isinstance(X, np.ndarray):
            X_loc = np.asarray(X)

        dtype = get_dtype(X_loc)
        if dtype not in [np.float32, np.float64]:
            dtype = np.float64
            X_loc = X_loc.astype(dtype, copy=self.copy_X)

        y_loc = np.asarray(y_loc).astype(dtype=dtype)

        # Finiteness is checked in the sklearnex wrapper
        X_loc, y_loc = _check_X_y(
            X_loc, y_loc, force_all_finite=False, accept_2d_y=True)

        self.n_features_in_ = _num_features(X_loc, fallback_1d=True)

        X_loc, y_loc = _convert_to_supported(policy, X_loc, y_loc)
        params = self._get_onedal_params(get_dtype(X_loc))
        X_table, y_table = to_table(X_loc, y_loc)

        result = module.train(policy, params, X_table, y_table)

        self._onedal_model = result.model

        coeff = from_table(result.model.packed_coefficients)
        self.coef_, self.intercept_ = coeff[:, 1:], coeff[:, 0]

        if self.coef_.shape[0] == 1 and y_loc.ndim == 1:
            self.coef_ = self.coef_.ravel()
            self.intercept_ = self.intercept_[0]

        return self

    def _create_model(self, module, policy):
        m = module.model()

        coefficients = self.coef_
        dtype = get_dtype(coefficients)
        if not isinstance(coefficients, np.ndarray):
            coefficients = np.asarray(coefficients, dtype=dtype)

        if coefficients.ndim == 2:
            n_features_in = coefficients.shape[1]
            n_targets_in = coefficients.shape[0]
        else:
            n_features_in = coefficients.size
            n_targets_in = 1

        intercept = self.intercept_
        if isinstance(intercept, Number):
            assert n_targets_in == 1
        else:
            if not isinstance(intercept, np.ndarray):
                intercept = np.asarray(intercept, dtype=dtype)
            assert n_targets_in == intercept.size

        intercept = _check_array(intercept, dtype=[np.float64, np.float32],
                                 force_all_finite=True, ensure_2d=False)
        coefficients = _check_array(
            coefficients,
            dtype=[
                np.float64,
                np.float32],
            force_all_finite=True, ensure_2d=False)

        coefficients, intercept = make2d(coefficients), make2d(intercept)
        coefficients = coefficients.T if n_targets_in == 1 else coefficients

        assert coefficients.shape == (n_targets_in, n_features_in)
        assert intercept.shape == (n_targets_in, 1)

        desired_shape = (n_targets_in, n_features_in + 1)
        packed_coefficients = np.zeros(desired_shape, dtype=dtype)

        packed_coefficients[:, 1:] = coefficients
        if self.fit_intercept:
            packed_coefficients[:, 0][:, np.newaxis] = intercept

        packed_coefficients = _convert_to_supported(policy, packed_coefficients)

        m.packed_coefficients = to_table(packed_coefficients)

        self._onedal_model = m

        return m

    def _predict(self, X, module, queue):
        _check_is_fitted(self)

        policy = self._get_policy(queue, X)

        if isinstance(X, np.ndarray):
            X_loc = np.asarray(X)
        else:
            X_loc = X

        # Finiteness is checked in the sklearnex wrapper
        X_loc = _check_array(X_loc, dtype=[np.float64, np.float32],
                             force_all_finite=False, ensure_2d=False)
        _check_n_features(self, X_loc, False)

        if hasattr(self, '_onedal_model'):
            model = self._onedal_model
        else:
            model = self._create_model(module, policy)

        X_loc = make2d(X_loc)
        X_loc = _convert_to_supported(policy, X_loc)
        params = self._get_onedal_params(get_dtype(X_loc))

        X_table = to_table(X_loc)
        result = module.infer(policy, params, model, X_table)
        y = from_table(result.responses)

        if not isinstance(self.coef_, np.ndarray):
            coefficients = np.asarray(self.coef_)
        else:
            coefficients = self.coef_

        if y.shape[1] == 1 and coefficients.ndim == 1:
            return y.ravel()
        else:
            return y


class LinearRegression(RegressorMixin, BaseLinearRegression):
    """
    Linear Regression oneDAL implementation.
    """

    def __init__(
            self,
            fit_intercept=True,
            copy_X=False,
            *,
            algorithm='norm_eq',
            **kwargs):
        super().__init__(fit_intercept=fit_intercept, copy_X=copy_X, algorithm=algorithm)

    def fit(self, X, y, queue=None):
        return super()._fit(X, y, _backend.linear_model.regression, queue)

    def predict(self, X, queue=None):
        y = super()._predict(X, _backend.linear_model.regression, queue)
        return y



#===============================================================================
# Copyright 2014 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

# daal4py quantiles example for shared memory systems

import daal4py as d4p
import numpy as np

# let's try to use pandas' fast csv reader
try:
    import pandas

    def read_csv(f, c=None, t=np.float64):
        return pandas.read_csv(f, usecols=c, delimiter=',', header=None, dtype=t)
except ImportError:
    # fall back to numpy loadtxt
    def read_csv(f, c=None, t=np.float64):
        return np.loadtxt(f, usecols=c, delimiter=',', ndmin=2)


def main(readcsv=read_csv, method='defaultDense'):
    infile = "./data/batch/quantiles.csv"

    # configure a quantiles object
    algo = d4p.quantiles()

    # let's provide a file directly, not a table/array
    result1 = algo.compute(infile)

    # We can also load the data ourselfs and provide the numpy array
    data = readcsv(infile)
    result2 = algo.compute(data)

    # QUANTILES result objects provide quantiles
    assert np.allclose(result1.quantiles, result2.quantiles)

    return result1


if __name__ == "__main__":
    result = main()
    print("Quantiles:\n", result.quantiles)
    print('All looks good!')



# ===============================================================================
# Copyright 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============================================================================

import numpy as np
from mpi4py import MPI
import dpctl
import dpctl.tensor as dpt
from sklearnex.spmd.decomposition import PCA


def get_data(data_seed):
    ns, nf = 300, 30
    drng = np.random.default_rng(data_seed)
    X = drng.random(size=(ns, nf))
    return X


q = dpctl.SyclQueue("gpu")
comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()

X = get_data(rank)
dpt_X = dpt.asarray(X, usm_type="device", sycl_queue=q)

pca = PCA(n_components=2).fit(dpt_X)

print(f"Singular values on rank {rank}:\n", pca.singular_values_)
print(f"Explained variance Ratio on rank {rank}:\n", pca.explained_variance_ratio_)



#===============================================================================
# Copyright 2014 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

from daal4py.sklearn._utils import set_idp_sklearn_verbose
from ..neighbors import KNeighborsRegressor as KNeighborsRegressor_daal4py
from ..neighbors import NearestNeighbors as NearestNeighbors_daal4py
from ..neighbors import KNeighborsClassifier as KNeighborsClassifier_daal4py
from ..model_selection import _daal_train_test_split
from ..utils.validation import _daal_assert_all_finite
from ..svm.svm import SVC as SVC_daal4py
from ..ensemble._forest import RandomForestClassifier as RandomForestClassifier_daal4py
from ..ensemble._forest import RandomForestRegressor as RandomForestRegressor_daal4py
from ..metrics import _daal_roc_auc_score
from ..metrics import daal_pairwise_distances
from ..cluster.k_means import KMeans as KMeans_daal4py
from ..cluster.dbscan import DBSCAN as DBSCAN_daal4py
from ..linear_model.coordinate_descent import Lasso as Lasso_daal4py
from ..linear_model.coordinate_descent import ElasticNet as ElasticNet_daal4py
from ..linear_model.linear import LinearRegression as LinearRegression_daal4py
from ..linear_model.ridge import Ridge as Ridge_daal4py
from ..linear_model.logistic_path import LogisticRegression as LogisticRegression_daal4py
from ..linear_model.logistic_path import logistic_regression_path as \
    daal_optimized_logistic_path
from ..decomposition._pca import PCA as PCA_daal4py
from ..manifold import TSNE as TSNE_daal4py
from sklearn import model_selection
from sklearn import metrics
from sklearn.utils import validation
import sys
from functools import lru_cache

import sklearn.cluster as cluster_module
import sklearn.ensemble as ensemble_module
import sklearn.svm as svm_module
import sklearn.linear_model._logistic as logistic_module
import sklearn.neighbors as neighbors_module
import sklearn.decomposition as decomposition_module
import sklearn.linear_model as linear_model_module
import sklearn.manifold as manifold_module

import warnings


@lru_cache(maxsize=None)
def _get_map_of_algorithms():
    mapping = {
        'pca': [[(decomposition_module, 'PCA', PCA_daal4py), None]],
        'kmeans': [[(cluster_module, 'KMeans', KMeans_daal4py), None]],
        'dbscan': [[(cluster_module, 'DBSCAN', DBSCAN_daal4py), None]],
        'distances': [[(metrics, 'pairwise_distances', daal_pairwise_distances), None]],
        'linear': [[(linear_model_module, 'LinearRegression',
                     LinearRegression_daal4py), None]],
        'ridge': [[(linear_model_module, 'Ridge', Ridge_daal4py), None]],
        'elasticnet': [[(linear_model_module, 'ElasticNet', ElasticNet_daal4py), None]],
        'lasso': [[(linear_model_module, 'Lasso', Lasso_daal4py), None]],
        'svm': [[(svm_module, 'SVC', SVC_daal4py), None]],
        'logistic': [[(logistic_module, '_logistic_regression_path',
                       daal_optimized_logistic_path), None]],
        'log_reg': [[(linear_model_module, 'LogisticRegression',
                    LogisticRegression_daal4py), None]],
        'knn_classifier': [[(neighbors_module, 'KNeighborsClassifier',
                             KNeighborsClassifier_daal4py), None]],
        'nearest_neighbors': [[(neighbors_module, 'NearestNeighbors',
                                NearestNeighbors_daal4py), None]],
        'knn_regressor': [[(neighbors_module, 'KNeighborsRegressor',
                            KNeighborsRegressor_daal4py), None]],
        'random_forest_classifier': [[(ensemble_module, 'RandomForestClassifier',
                                       RandomForestClassifier_daal4py), None]],
        'random_forest_regressor': [[(ensemble_module, 'RandomForestRegressor',
                                      RandomForestRegressor_daal4py), None]],
        'train_test_split': [[(model_selection, 'train_test_split',
                               _daal_train_test_split), None]],
        'fin_check': [[(validation, '_assert_all_finite',
                        _daal_assert_all_finite), None]],
        'roc_auc_score': [[(metrics, 'roc_auc_score',
                           _daal_roc_auc_score), None]],
        'tsne': [[(manifold_module, 'TSNE', TSNE_daal4py), None]],
    }
    mapping['svc'] = mapping['svm']
    mapping['logisticregression'] = mapping['log_reg']
    mapping['kneighborsclassifier'] = mapping['knn_classifier']
    mapping['nearestneighbors'] = mapping['nearest_neighbors']
    mapping['kneighborsregressor'] = mapping['knn_regressor']
    mapping['randomrorestclassifier'] = mapping['random_forest_classifier']
    mapping['randomforestregressor'] = mapping['random_forest_regressor']
    return mapping


def do_patch(name, get_map=_get_map_of_algorithms):
    lname = name.lower()
    if lname in get_map():
        for descriptor in get_map()[lname]:
            which, what, replacer = descriptor[0]
            if descriptor[1] is None:
                descriptor[1] = getattr(which, what, None)
            setattr(which, what, replacer)
    else:
        raise ValueError("Has no patch for: " + name)


def do_unpatch(name, get_map=_get_map_of_algorithms):
    lname = name.lower()
    if lname in get_map():
        for descriptor in get_map()[lname]:
            if descriptor[1] is not None:
                which, what, _ = descriptor[0]
                setattr(which, what, descriptor[1])
    else:
        raise ValueError("Has no patch for: " + name)


def enable(name=None, verbose=True, deprecation=True, get_map=_get_map_of_algorithms):
    if name is not None:
        do_patch(name, get_map)
    else:
        for key in get_map():
            do_patch(key, get_map)
    if deprecation:
        set_idp_sklearn_verbose()
        warnings.warn_explicit("\nScikit-learn patching with daal4py is deprecated "
                               "and will be removed in the future.\n"
                               "Use Intel(R) Extension "
                               "for Scikit-learn* module instead "
                               "(pip install scikit-learn-intelex).\n"
                               "To enable patching, please use one of the "
                               "following options:\n"
                               "1) From the command line:\n"
                               "    python -m sklearnex <your_script>\n"
                               "2) From your script:\n"
                               "    from sklearnex import patch_sklearn\n"
                               "    patch_sklearn()",
                               FutureWarning, "dispatcher.py", 151)
    if verbose and deprecation and sys.stderr is not None:
        sys.stderr.write(
            "Intel(R) oneAPI Data Analytics Library solvers for sklearn enabled: "
            "https://intelpython.github.io/daal4py/sklearn.html\n")


def disable(name=None, get_map=_get_map_of_algorithms):
    if name is not None:
        do_unpatch(name, get_map)
    else:
        for key in get_map():
            do_unpatch(key, get_map)
        get_map.cache_clear()


def _is_enabled(name, get_map=_get_map_of_algorithms):
    lname = name.lower()
    if lname in get_map():
        enabled = True
        for descriptor in get_map()[lname]:
            which, what, replacer = descriptor[0]
            enabled = enabled and getattr(which, what) == replacer
        return enabled
    else:
        raise ValueError("Has no patch for: " + name)


def patch_is_enabled(name=None, get_map=_get_map_of_algorithms, return_map=False):
    if name is not None:
        return _is_enabled(name, get_map)
    else:
        if return_map:
            enabled = {}
            for key in get_map():
                enabled[key] = _is_enabled(key, get_map)
        else:
            enabled = True
            for key in get_map():
                enabled = enabled and _is_enabled(key, get_map)
        return enabled


def _patch_names():
    return list(_get_map_of_algorithms().keys())



# ===============================================================================
# Copyright 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============================================================================

from daal4py.sklearn._utils import (
    daal_check_version, sklearn_check_version)
from sklearn.ensemble import BaseEnsemble
from abc import ABCMeta, abstractmethod
import numbers
from numbers import Number
import warnings
from sklearn.exceptions import DataConversionWarning
from sklearn.utils import (
    check_random_state,
    compute_sample_weight,
    check_array,
    deprecated)
from sklearn.utils.validation import (
    check_is_fitted,
    check_consistent_length,
    _num_samples)
from math import ceil

import numpy as np
from scipy import sparse as sp
from ..datatypes import (
    _validate_targets,
    _check_X_y,
    _check_array,
    _column_or_1d,
    _check_n_features,
    _convert_to_supported
)

from ..common._mixin import ClassifierMixin, RegressorMixin
from ..common._policy import _get_policy
from ..common._estimator_checks import _check_is_fitted
from ..datatypes._data_conversion import from_table, to_table
from onedal import _backend

from sklearn.tree import DecisionTreeClassifier


class BaseForest(BaseEnsemble, metaclass=ABCMeta):
    @abstractmethod
    def __init__(
            self,
            n_estimators,
            criterion,
            max_depth,
            min_samples_split,
            min_samples_leaf,
            min_weight_fraction_leaf,
            max_features,
            max_leaf_nodes,
            min_impurity_decrease,
            min_impurity_split,
            bootstrap,
            oob_score,
            random_state,
            warm_start,
            class_weight,
            ccp_alpha,
            max_samples,
            max_bins,
            min_bin_size,
            infer_mode,
            splitter_mode,
            voting_mode,
            error_metric_mode,
            variable_importance_mode,
            algorithm,
            **kwargs):
        self.n_estimators = n_estimators
        self.bootstrap = bootstrap
        self.oob_score = oob_score
        self.random_state = random_state
        self.warm_start = warm_start
        self.class_weight = class_weight
        self.max_samples = max_samples
        self.criterion = criterion
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_features = max_features
        self.max_leaf_nodes = max_leaf_nodes
        self.min_impurity_decrease = min_impurity_decrease
        self.min_impurity_split = min_impurity_split
        self.ccp_alpha = ccp_alpha
        self.max_bins = max_bins
        self.min_bin_size = min_bin_size
        self.infer_mode = infer_mode
        self.splitter_mode = splitter_mode
        self.voting_mode = voting_mode
        self.error_metric_mode = error_metric_mode
        self.variable_importance_mode = variable_importance_mode
        self.algorithm = algorithm

    def _to_absolute_max_features(self, max_features, n_features,
                                  is_classification=False):
        if max_features is None:
            return n_features
        if isinstance(max_features, str):
            if max_features == "auto":
                if not sklearn_check_version('1.3'):
                    if sklearn_check_version('1.1'):
                        warnings.warn(
                            "`max_features='auto'` has been deprecated in 1.1 "
                            "and will be removed in 1.3. To keep the past behaviour, "
                            "explicitly set `max_features=1.0` or remove this "
                            "parameter as it is also the default value for "
                            "RandomForestRegressors and ExtraTreesRegressors.",
                            FutureWarning, )
                    return max(1, int(np.sqrt(n_features))
                               ) if is_classification else n_features
            if max_features == 'sqrt':
                return max(1, int(np.sqrt(n_features)))
            if max_features == "log2":
                return max(1, int(np.log2(n_features)))
            allowed_string_values = '"sqrt" or "log2"' if sklearn_check_version(
                '1.3') else '"auto", "sqrt" or "log2"'
            raise ValueError(
                'Invalid value for max_features. Allowed string '
                f'values are {allowed_string_values}.')
        if isinstance(max_features, (numbers.Integral, np.integer)):
            return max_features
        if max_features > 0.0:
            return max(1, int(max_features * n_features))
        return 0

    def _get_observations_per_tree_fraction(self, n_samples, max_samples):
        if max_samples is None:
            return 1.

        if isinstance(max_samples, numbers.Integral):
            if not sklearn_check_version('1.2'):
                if not (1 <= max_samples <= n_samples):
                    msg = "`max_samples` must be in range 1 to {} but got value {}"
                    raise ValueError(msg.format(n_samples, max_samples))
            else:
                if max_samples > n_samples:
                    msg = "`max_samples` must be <= n_samples={} but got value {}"
                    raise ValueError(msg.format(n_samples, max_samples))
            return float(max_samples / n_samples)

        if isinstance(max_samples, numbers.Real):
            if sklearn_check_version('1.2'):
                pass
            elif sklearn_check_version('1.0'):
                if not (0 < float(max_samples) <= 1):
                    msg = "`max_samples` must be in range (0.0, 1.0] but got value {}"
                    raise ValueError(msg.format(max_samples))
            else:
                if not (0 < float(max_samples) < 1):
                    msg = "`max_samples` must be in range (0, 1) but got value {}"
                    raise ValueError(msg.format(max_samples))
            return float(max_samples)

        msg = "`max_samples` should be int or float, but got type '{}'"
        raise TypeError(msg.format(type(max_samples)))

    def _get_onedal_params(self, data):
        n_samples, n_features = data.shape
        features_per_node = self._to_absolute_max_features(
            self.max_features, n_features, self.is_classification)

        observations_per_tree_fraction = self._get_observations_per_tree_fraction(
            n_samples=n_samples, max_samples=self.max_samples)
        observations_per_tree_fraction = observations_per_tree_fraction if bool(
            self.bootstrap) else 1.

        if not self.bootstrap and self.max_samples is not None:
            raise ValueError(
                "`max_sample` cannot be set if `bootstrap=False`. "
                "Either switch to `bootstrap=True` or set "
                "`max_sample=None`."
            )
        if not self.bootstrap and self.oob_score:
            raise ValueError("Out of bag estimation only available"
                             " if bootstrap=True")

        min_observations_in_leaf_node = (
            self.min_samples_leaf if isinstance(
                self.min_samples_leaf,
                numbers.Integral) else int(
                ceil(
                    self.min_samples_leaf * n_samples)))

        min_observations_in_split_node = (
            self.min_samples_split if isinstance(
                self.min_samples_split,
                numbers.Integral) else int(
                ceil(
                    self.min_samples_split * n_samples)))

        onedal_params = {
            'fptype': 'float' if data.dtype == np.float32 else 'double',
            'method': self.algorithm,
            'infer_mode': self.infer_mode,
            'voting_mode': self.voting_mode,
            'observations_per_tree_fraction': observations_per_tree_fraction,
            'impurity_threshold': float(
                0.0 if self.min_impurity_split is None else self.min_impurity_split),
            'min_weight_fraction_in_leaf_node': self.min_weight_fraction_leaf,
            'min_impurity_decrease_in_split_node': self.min_impurity_decrease,
            'tree_count': int(self.n_estimators),
            'features_per_node': features_per_node,
            'max_tree_depth': int(0 if self.max_depth is None else self.max_depth),
            'min_observations_in_leaf_node': min_observations_in_leaf_node,
            'min_observations_in_split_node': min_observations_in_split_node,
            'max_leaf_nodes': (0 if self.max_leaf_nodes is None else self.max_leaf_nodes),
            'max_bins': self.max_bins,
            'min_bin_size': self.min_bin_size,
            'memory_saving_mode': False,
            'bootstrap': bool(self.bootstrap),
            'error_metric_mode': self.error_metric_mode,
            'variable_importance_mode': self.variable_importance_mode,
        }
        if self.is_classification:
            onedal_params['class_count'] = 0 if self.classes_ is None else len(
                self.classes_)
        if daal_check_version((2023, 'P', 101)):
            onedal_params['splitter_mode'] = self.splitter_mode
        return onedal_params

    def _check_parameters(self):
        if isinstance(self.min_samples_leaf, numbers.Integral):
            if not 1 <= self.min_samples_leaf:
                raise ValueError("min_samples_leaf must be at least 1 "
                                 "or in (0, 0.5], got %s"
                                 % self.min_samples_leaf)
        else:  # float
            if not 0. < self.min_samples_leaf <= 0.5:
                raise ValueError("min_samples_leaf must be at least 1 "
                                 "or in (0, 0.5], got %s"
                                 % self.min_samples_leaf)
        if isinstance(self.min_samples_split, numbers.Integral):
            if not 2 <= self.min_samples_split:
                raise ValueError("min_samples_split must be an integer "
                                 "greater than 1 or a float in (0.0, 1.0]; "
                                 "got the integer %s"
                                 % self.min_samples_split)
        else:  # float
            if not 0. < self.min_samples_split <= 1.:
                raise ValueError("min_samples_split must be an integer "
                                 "greater than 1 or a float in (0.0, 1.0]; "
                                 "got the float %s"
                                 % self.min_samples_split)
        if not 0 <= self.min_weight_fraction_leaf <= 0.5:
            raise ValueError("min_weight_fraction_leaf must in [0, 0.5]")
        if self.min_impurity_split is not None:
            warnings.warn("The min_impurity_split parameter is deprecated. "
                          "Its default value has changed from 1e-7 to 0 in "
                          "version 0.23, and it will be removed in 0.25. "
                          "Use the min_impurity_decrease parameter instead.",
                          FutureWarning)

            if self.min_impurity_split < 0.:
                raise ValueError("min_impurity_split must be greater than "
                                 "or equal to 0")
        if self.min_impurity_decrease < 0.:
            raise ValueError("min_impurity_decrease must be greater than "
                             "or equal to 0")
        if self.max_leaf_nodes is not None:
            if not isinstance(self.max_leaf_nodes, numbers.Integral):
                raise ValueError(
                    "max_leaf_nodes must be integral number but was "
                    "%r" %
                    self.max_leaf_nodes)
            if self.max_leaf_nodes < 2:
                raise ValueError(
                    ("max_leaf_nodes {0} must be either None "
                     "or larger than 1").format(
                        self.max_leaf_nodes))
        if isinstance(self.max_bins, numbers.Integral):
            if not 2 <= self.max_bins:
                raise ValueError("max_bins must be at least 2, got %s"
                                 % self.max_bins)
        else:
            raise ValueError("max_bins must be integral number but was "
                             "%r" % self.max_bins)
        if isinstance(self.min_bin_size, numbers.Integral):
            if not 1 <= self.min_bin_size:
                raise ValueError("min_bin_size must be at least 1, got %s"
                                 % self.min_bin_size)
        else:
            raise ValueError("min_bin_size must be integral number but was "
                             "%r" % self.min_bin_size)

    def _validate_targets(self, y, dtype):
        self.class_weight_ = None
        self.classes_ = None
        return _column_or_1d(y, warn=True).astype(dtype, copy=False)

    def _get_sample_weight(self, X, y, sample_weight):
        n_samples = X.shape[0]
        dtype = X.dtype
        if n_samples == 1:
            raise ValueError("n_samples=1")

        sample_weight = np.asarray([]
                                   if sample_weight is None
                                   else sample_weight, dtype=dtype)
        sample_weight = sample_weight.ravel()

        sample_weight_count = sample_weight.shape[0]
        if sample_weight_count != 0 and sample_weight_count != n_samples:
            raise ValueError("sample_weight and X have incompatible shapes: "
                             "%r vs %r\n"
                             "Note: Sparse matrices cannot be indexed w/"
                             "boolean masks (use `indices=True` in CV)."
                             % (len(sample_weight), X.shape))

        if sample_weight_count == 0:
            sample_weight = np.ones(n_samples, dtype=dtype)
        elif isinstance(sample_weight, Number):
            sample_weight = np.full(n_samples, sample_weight, dtype=dtype)
        else:
            sample_weight = _check_array(
                sample_weight, accept_sparse=False, ensure_2d=False,
                dtype=dtype, order="C"
            )
            if sample_weight.ndim != 1:
                raise ValueError("Sample weights must be 1D array or scalar")

            if sample_weight.shape != (n_samples,):
                raise ValueError("sample_weight.shape == {}, expected {}!"
                                 .format(sample_weight.shape, (n_samples,)))
        return sample_weight

    def _get_policy(self, queue, *data):
        return _get_policy(queue, *data)

    def _fit(self, X, y, sample_weight, module, queue):
        X, y = _check_X_y(
            X, y, dtype=[np.float64, np.float32],
            force_all_finite=True, accept_sparse='csr')
        y = self._validate_targets(y, X.dtype)
        sample_weight = self._get_sample_weight(X, y, sample_weight)

        self.n_features_in_ = X.shape[1]
        if not sklearn_check_version('1.0'):
            self.n_features_ = self.n_features_in_
        policy = self._get_policy(queue, X, y, sample_weight)

        X, y, sample_weight = _convert_to_supported(policy, X, y, sample_weight)
        params = self._get_onedal_params(X)
        train_result = module.train(
            policy, params, *to_table(X, y, sample_weight))
        self._onedal_model = train_result.model

        if self.oob_score:
            if self.is_classification:
                self.oob_score_ = from_table(train_result.oob_err_accuracy)[0, 0]
                self.oob_prediction_ = from_table(
                    train_result.oob_err_decision_function)
            else:
                self.oob_score_ = from_table(train_result.oob_err_r2)[0, 0]
                self.oob_prediction_ = from_table(
                    train_result.oob_err_prediction).reshape(-1)
            if np.any(self.oob_prediction_ == 0):
                warnings.warn(
                    "Some inputs do not have OOB scores. This probably means "
                    "too few trees were used to compute any reliable OOB "
                    "estimates.",
                    UserWarning,
                )

        return self

    def _create_model(self, module):
        # TODO:
        # upate error msg.
        raise NotImplementedError('Creating model is not supported.')

    def _predict(self, X, module, queue):
        _check_is_fitted(self)
        X = _check_array(X, dtype=[np.float64, np.float32],
                         force_all_finite=True, accept_sparse=False)
        _check_n_features(self, X, False)
        policy = self._get_policy(queue, X)

        model = self._onedal_model
        X = _convert_to_supported(policy, X)
        params = self._get_onedal_params(X)
        result = module.infer(policy, params, model, to_table(X))
        y = from_table(result.responses)
        return y

    def _predict_proba(self, X, module, queue):
        _check_is_fitted(self)
        X = _check_array(X, dtype=[np.float64, np.float32],
                         force_all_finite=True, accept_sparse=False)
        _check_n_features(self, X, False)
        policy = self._get_policy(queue, X)
        X = _convert_to_supported(policy, X)
        params = self._get_onedal_params(X)
        params['infer_mode'] = 'class_probabilities'

        model = self._onedal_model
        result = module.infer(policy, params, model, to_table(X))
        y = from_table(result.probabilities)
        return y


class RandomForestClassifier(ClassifierMixin, BaseForest, metaclass=ABCMeta):
    def __init__(self,
                 n_estimators=100,
                 criterion="gini",
                 max_depth=None,
                 min_samples_split=2,
                 min_samples_leaf=1,
                 min_weight_fraction_leaf=0.,
                 max_features="auto",
                 max_leaf_nodes=None,
                 min_impurity_decrease=0.,
                 min_impurity_split=None,
                 bootstrap=True,
                 oob_score=False,
                 random_state=None,
                 warm_start=False,
                 class_weight=None,
                 ccp_alpha=0.0,
                 max_samples=None,
                 max_bins=256,
                 min_bin_size=1,
                 infer_mode='class_responses',
                 splitter_mode='best',
                 voting_mode='weighted',
                 error_metric_mode='none',
                 variable_importance_mode='none',
                 algorithm='hist',
                 **kwargs):
        super().__init__(
            n_estimators=n_estimators,
            criterion=criterion,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
            max_features=max_features,
            max_leaf_nodes=max_leaf_nodes,
            min_impurity_decrease=min_impurity_decrease,
            min_impurity_split=min_impurity_split,
            bootstrap=bootstrap,
            oob_score=oob_score,
            random_state=random_state,
            warm_start=warm_start,
            class_weight=class_weight,
            ccp_alpha=ccp_alpha,
            max_samples=max_samples,
            max_bins=max_bins,
            min_bin_size=min_bin_size,
            infer_mode=infer_mode,
            splitter_mode=splitter_mode,
            voting_mode=voting_mode,
            error_metric_mode=error_metric_mode,
            variable_importance_mode=variable_importance_mode,
            algorithm=algorithm)
        self.is_classification = True

    def _validate_targets(self, y, dtype):
        y, self.class_weight_, self.classes_ = _validate_targets(
            y, self.class_weight, dtype)

        # Decapsulate classes_ attributes
        # TODO:
        # align with `n_classes_` and `classes_` attr with daal4py implementations.
        # if hasattr(self, "classes_"):
        #    self.n_classes_ = self.classes_
        return y

    def fit(self, X, y, sample_weight=None, queue=None):
        return self._fit(X, y, sample_weight,
                         _backend.decision_forest.classification, queue)

    def predict(self, X, queue=None):
        pred = super()._predict(X, _backend.decision_forest.classification, queue)

        return np.take(
            self.classes_,
            pred.ravel().astype(
                np.int64,
                casting='unsafe'))

    def predict_proba(self, X, queue=None):
        return super()._predict_proba(X, _backend.decision_forest.classification, queue)


class RandomForestRegressor(RegressorMixin, BaseForest, metaclass=ABCMeta):
    def __init__(self,
                 n_estimators=100,
                 criterion="squared_error",
                 max_depth=None,
                 min_samples_split=2,
                 min_samples_leaf=1,
                 min_weight_fraction_leaf=0.,
                 max_features="auto",
                 max_leaf_nodes=None,
                 min_impurity_decrease=0.,
                 min_impurity_split=None,
                 bootstrap=True,
                 oob_score=False,
                 random_state=None,
                 warm_start=False,
                 class_weight=None,
                 ccp_alpha=0.0,
                 max_samples=None,
                 max_bins=256,
                 min_bin_size=1,
                 infer_mode='class_responses',
                 splitter_mode='best',
                 voting_mode='weighted',
                 error_metric_mode='none',
                 variable_importance_mode='none',
                 algorithm='hist',
                 **kwargs):
        super().__init__(
            n_estimators=n_estimators,
            criterion=criterion,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
            max_features=max_features,
            max_leaf_nodes=max_leaf_nodes,
            min_impurity_decrease=min_impurity_decrease,
            min_impurity_split=min_impurity_split,
            bootstrap=bootstrap,
            oob_score=oob_score,
            random_state=random_state,
            warm_start=warm_start,
            class_weight=class_weight,
            ccp_alpha=ccp_alpha,
            max_samples=max_samples,
            max_bins=max_bins,
            min_bin_size=min_bin_size,
            infer_mode=infer_mode,
            splitter_mode=splitter_mode,
            voting_mode=voting_mode,
            error_metric_mode=error_metric_mode,
            variable_importance_mode=variable_importance_mode,
            algorithm=algorithm)
        self.is_classification = False

    def fit(self, X, y, sample_weight=None, queue=None):
        if sample_weight is not None:
            if hasattr(sample_weight, '__array__'):
                sample_weight[sample_weight == 0.0] = 1.0
            sample_weight = [sample_weight]
        return super()._fit(X, y, sample_weight,
                            _backend.decision_forest.regression, queue)

    def predict(self, X, queue=None):
        return super()._predict(X, _backend.decision_forest.regression, queue).ravel()



# ===============================================================================
# Copyright 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============================================================================

from daal4py.sklearn._utils import daal_check_version, sklearn_check_version

if daal_check_version((2023, 'P', 100)):
    import pytest
    import numpy as np
    from numpy.testing import assert_allclose

    from onedal.basic_statistics import BasicStatistics
    from onedal.tests.utils._device_selection import get_queues

    options_and_tests = [
        ("sum", np.sum, (1e-5, 1e-7)),
        ("min", np.min, (1e-5, 1e-7)),
        ("max", np.max, (1e-5, 1e-7)),
        ("mean", np.mean, (1e-5, 1e-7)),
        ("standard_deviation", np.std, (3e-5, 3e-5))
    ]

    @pytest.mark.parametrize('queue', get_queues())
    @pytest.mark.parametrize('dtype', [np.float32, np.float64])
    def test_basic_uniform(queue, dtype):
        seed = 42
        s_count, f_count = 70000, 29

        gen = np.random.default_rng(seed)
        data = gen.uniform(low=-0.5, high=+0.6,
                           size=(s_count, f_count))
        data = data.astype(dtype=dtype)

        alg = BasicStatistics(result_options="mean")
        res = alg.compute(data, queue=queue)

        res_mean = res["mean"]
        gtr_mean = np.mean(data, axis=0)
        tol = 1e-5 if res_mean.dtype == np.float32 else 1e-7
        assert_allclose(gtr_mean, res_mean, rtol=tol)

    @pytest.mark.parametrize('queue', get_queues())
    @pytest.mark.parametrize('option', options_and_tests)
    @pytest.mark.parametrize('dtype', [np.float32, np.float64])
    def test_option_uniform(queue, option, dtype):
        seed = 77
        s_count, f_count = 19999, 31

        result_option, function, tols = option
        fp32tol, fp64tol = tols

        gen = np.random.default_rng(seed)
        data = gen.uniform(low=-0.3, high=+0.7,
                           size=(s_count, f_count))
        data = data.astype(dtype=dtype)

        alg = BasicStatistics(result_options=result_option)
        res = alg.compute(data, queue=queue)

        res, gtr = res[result_option], function(data, axis=0)

        tol = fp32tol if res.dtype == np.float32 else fp64tol
        assert_allclose(gtr, res, rtol=tol)

    @pytest.mark.parametrize('queue', get_queues())
    @pytest.mark.parametrize('option', options_and_tests)
    @pytest.mark.parametrize('dtype', [np.float32, np.float64])
    def test_option_weighted(queue, option, dtype):
        seed = 999
        s_count, f_count = 1024, 127

        result_option, function, tols = option
        fp32tol, fp64tol = tols
        fp32tol, fp64tol = 30 * fp32tol, 50 * fp64tol

        gen = np.random.default_rng(seed)
        data = gen.uniform(low=-5.0, high=+9.0,
                           size=(s_count, f_count))
        weights = gen.uniform(low=-0.5, high=+1.0,
                              size=s_count)

        data = data.astype(dtype=dtype)
        weights = weights.astype(dtype=dtype)

        alg = BasicStatistics(result_options=result_option)
        res = alg.compute(data, weights, queue=queue)

        weighted = np.diag(weights) @ data
        res, gtr = res[result_option], function(weighted, axis=0)

        tol = fp32tol if res.dtype == np.float32 else fp64tol
        assert_allclose(gtr, res, rtol=tol)



#===============================================================================
# Copyright 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

import numpy as np
from sklearn.metrics import accuracy_score
from warnings import warn
from mpi4py import MPI
import dpctl
import dpctl.tensor as dpt
from sklearnex.spmd.neighbors import KNeighborsClassifier


def generate_X_y(par, seed):
    ns, nf = par['ns'], par['nf']

    drng = np.random.default_rng(seed)
    data = drng.uniform(-1, 1, size=(ns, nf))
    resp = (data > 0) @ (2 ** np.arange(nf))

    return data, resp


comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()

if dpctl.has_gpu_devices:
    q = dpctl.SyclQueue("gpu")
else:
    raise RuntimeError('GPU devices unavailable. Currently, '
          'SPMD execution mode is implemented only for this device type.')

params_train = {'ns': 100000, 'nf': 8}
params_test = {'ns': 100, 'nf': 8}

X_train, y_train = generate_X_y(params_train, rank)
X_test, y_test = generate_X_y(params_test, rank + 99)

dpt_X_train = dpt.asarray(X_train, usm_type="device", sycl_queue=q)
dpt_y_train = dpt.asarray(y_train, usm_type="device", sycl_queue=q)
dpt_X_test = dpt.asarray(X_test, usm_type="device", sycl_queue=q)
dpt_y_test = dpt.asarray(y_test, usm_type="device", sycl_queue=q)

model_spmd = KNeighborsClassifier(algorithm='brute',
                                  n_neighbors=20,
                                  weights='uniform',
                                  p=2,
                                  metric='minkowski')
model_spmd.fit(dpt_X_train, dpt_y_train)

y_predict = model_spmd.predict(dpt_X_test)

print("Brute Force Distributed kNN classification results:")
print("Ground truth (first 5 observations on rank {}):\n{}".format(rank, y_test[:5]))
print("Classification results (first 5 observations on rank {}):\n{}"
      .format(rank, dpt.to_numpy(y_predict)[:5]))
print("Accuracy for entire rank {} (256 classes): {}\n"
      .format(rank, accuracy_score(y_test, dpt.to_numpy(y_predict))))



#===============================================================================
# Copyright 2014 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

# daal4py normal distribution example for shared memory systems

import daal4py as d4p
import numpy as np


def main(readcsv=None, method='defaultDense'):

    # Create algorithm
    algorithm = d4p.distributions_normal(engine=d4p.engines_mt19937(seed=777))

    # Create array and fill with normal distribution
    data = np.zeros((1, 10))
    res = algorithm.compute(data)

    assert np.allclose(data, res.randomNumbers)
    assert np.allclose(
        data,
        [[
            -0.74104167, -0.13616829, -0.13679562, 2.40385531, -0.33556821,
            0.19041699, -0.61331181, 0.95958821, -0.42301092, 0.09460208
        ]]
    )

    return data


if __name__ == "__main__":
    res = main()
    print("\nNormal distribution output:", res)
    print("All looks good!")
